import logging

from cached_property import cached_property

from pymobiledevice3.exceptions import NoDeviceConnectedError, ConnectionFailedError
from pymobiledevice3.lockdown import list_devices
from pymobiledevice3.restore.restore_options import RestoreOptions
from pymobiledevice3.service_connection import ServiceConnection


class RestoredClient(object):
    DEFAULT_CLIENT_NAME = 'pyMobileDevice'
    SERVICE_PORT = 62078

    def __init__(self, udid=None, client_name=DEFAULT_CLIENT_NAME):
        udid = self._get_or_verify_udid(udid)
        self.logger = logging.getLogger(__name__)
        self.service = ServiceConnection.create(udid, self.SERVICE_PORT)
        self.label = client_name
        self.query_type = self.service.send_recv_plist({'Request': 'QueryType'})
        self.version = self.query_type.get('RestoreProtocolVersion')

        assert self.query_type.get('Type') == 'com.apple.mobile.restored', f'wrong query type: {self.query_type}'

    @staticmethod
    def _get_or_verify_udid(udid=None):
        available_udids = list_devices()
        if udid is None:
            if len(available_udids) == 0:
                raise NoDeviceConnectedError()
            return available_udids[0]
        else:
            if udid not in available_udids:
                raise ConnectionFailedError()
            return udid

    def query_value(self, key=None):
        req = {'Request': 'QueryValue', 'Label': self.label}

        if key:
            req['QueryKey'] = key

        return self.service.send_recv_plist(req)

    def start_restore(self, opts: RestoreOptions = None):
        req = {'Request': 'StartRestore', 'Label': self.label, 'RestoreProtocolVersion': self.version}

        if opts is not None:
            req['RestoreOptions'] = opts.to_dict()

        logging.debug(f'start_restore request: {req}')

        return self.service.send_plist(req)

    def reboot(self):
        return self.service.send_recv_plist({'Request': 'Reboot', 'Label': self.label})

    def send(self, message):
        self.service.send_plist(message)

    def recv(self):
        return self.service.recv_plist()

    @cached_property
    def hardware_info(self):
        return self.query_value('HardwareInfo')['HardwareInfo']


lpol_file = bytearray([0x30, 0x14, 0x16, 0x04, 0x49, 0x4d, 0x34, 0x50,
                       0x16, 0x04, 0x6c, 0x70, 0x6f, 0x6c, 0x16, 0x03,
                       0x31, 0x2e, 0x30, 0x04, 0x01, 0x00])

# extracted from ac2
PROGRESS_BAR_OPERATIONS = {
    11: 'CREATE_PARTITION_MAP',
    12: 'CREATE_FILESYSTEM',
    13: 'RESTORE_IMAGE',
    14: 'VERIFY_RESTORE',
    15: 'CHECK_FILESYSTEMS',
    16: 'MOUNT_FILESYSTEMS',
    17: 'FIXUP_VAR',
    18: 'FLASH_FIRMWARE',
    19: 'UPDATE_BASEBAND',
    20: 'SET_BOOT_STAGE',
    21: 'REBOOT_DEVICE',
    22: 'SHUTDOWN_DEVICE',
    23: 'TURN_ON_ACCESSORY_POWER',
    24: 'CLEAR_BOOTARGS',
    25: 'MODIFY_BOOTARGS',
    26: 'INSTALL_ROOT',
    27: 'INSTALL_KERNELCACHE',
    28: 'WAIT_FOR_NAND',
    29: 'UNMOUNT_FILESYSTEMS',
    30: 'SET_DATETIME',
    31: 'EXEC_IBOOT',
    32: 'FINALIZE_NAND_EPOCH_UPDATE',
    33: 'CHECK_INAPPR_BOOT_PARTITIONS',
    34: 'CREATE_FACTORY_RESTORE_MARKER',
    35: 'LOAD_FIRMWARE',
    36: 'REQUESTING_FUD_DATA',
    37: 'REMOVING_ACTIVATION_RECORD',
    38: 'CHECK_BATTERY_VOLTAGE',
    39: 'WAIT_BATTERY_CHARGE',
    40: 'CLOSE_MODEM_TICKETS',
    41: 'MIGRATE_DATA',
    42: 'WIPE_STORAGE_DEVICE',
    43: 'SEND_APPLE_LOGO',
    44: 'CHECK_LOGS',
    46: 'CLEAR_NVRAM',
    47: 'UPDATE_GAS_GAUGE',
    48: 'PREPARE_BASEBAND_UPDATE',
    49: 'BOOT_BASEBAND',
    50: 'CREATE_SYSTEM_KEYBAG',
    51: 'UPDATE_IR_MCU_FIRMWARE',
    52: 'RESIZE_SYSTEM_PARTITION',
    53: 'COLLECTING_UPDATER_OUTPUT',
    54: 'PAIR_STOCKHOLM',
    55: 'UPDATE_STOCKHOLM',
    56: 'UPDATE_SWDHID',
    57: 'CERTIFY_SEP',
    58: 'UPDATE_NAND_FIRMWARE',
    59: 'UPDATE_SE_FIRMWARE',
    60: 'UPDATE_SAVAGE',
    61: 'INSTALLING_DEVICETREE',
    62: 'CERTIFY_SAVAGE',
    63: 'SUBMITTING_PROVINFO',
    64: 'CERTIFY_YONKERS',
    65: 'UPDATE_ROSE',
    66: 'UPDATE_VERIDIAN',
    67: 'CREATING_PROTECTED_VOLUME',
    68: 'RESIZING_MAIN_FS_PARTITION',
    69: 'CREATING_RECOVERY_OS_VOLUME',
    70: 'INSTALLING_RECOVERY_OS_FILES',
    71: 'INSTALLING_RECOVERY_OS_IMAGE',
    74: 'REQUESTING_EAN_DATA',
    77: 'SEALING_SYSTEM_VOLUME',
}
SUPPORTED_DATA_TYPES = {
    'BasebandBootData': False,
    'BasebandData': False,
    'BasebandStackData': False,
    'BasebandUpdaterOutputData': False,
    'BuildIdentityDict': False,
    'BuildIdentityDictV2': False,
    'DataType': False,
    'DiagData': False,
    'EANData': False,
    'FDRMemoryCommit': False,
    'FDRTrustData': False,
    'FUDData': False,
    'FileData': False,
    'FileDataDone': False,
    'FirmwareUpdaterData': False,
    'GrapeFWData': False,
    'HPMFWData': False,
    'HostSystemTime': True,
    'KernelCache': False,
    'NORData': False,
    'NitrogenFWData': True,
    'OpalFWData': False,
    'OverlayRootDataCount': False,
    'OverlayRootDataForKey': True,
    'PeppyFWData': True,
    'PersonalizedBootObjectV3': False,
    'PersonalizedData': True,
    'ProvisioningData': False,
    'RamdiskFWData': True,
    'RecoveryOSASRImage': True,
    'RecoveryOSAppleLogo': True,
    'RecoveryOSDeviceTree': True,
    'RecoveryOSFileAssetImage': True,
    'RecoveryOSIBEC': True,
    'RecoveryOSIBootFWFilesImages': True,
    'RecoveryOSImage': True,
    'RecoveryOSKernelCache': True,
    'RecoveryOSLocalPolicy': True,
    'RecoveryOSOverlayRootDataCount': False,
    'RecoveryOSRootTicketData': True,
    'RecoveryOSStaticTrustCache': True,
    'RecoveryOSVersionData': True,
    'RootData': False,
    'RootTicket': False,
    'S3EOverride': False,
    'SourceBootObjectV3': False,
    'SourceBootObjectV4': False,
    'SsoServiceTicket': False,
    'StockholmPostflight': False,
    'SystemImageCanonicalMetadata': False,
    'SystemImageData': False,
    'SystemImageRootHash': False,
    'USBCFWData': False,
    'USBCOverride': False,
}

# extracted from ac2
SUPPORTED_MESSAGE_TYPES = {
    'BBUpdateStatusMsg': False,
    'CheckpointMsg': True,
    'DataRequestMsg': False,
    'FDRSubmit': True,
    'MsgType': False,
    'PreviousRestoreLogMsg': False,
    'ProgressMsg': False,
    'ProvisioningAck': False,
    'ProvisioningInfo': False,
    'ProvisioningStatusMsg': False,
    'ReceivedFinalStatusMsg': False,
    'RestoredCrash': True,
    'StatusMsg': False,
}



from distutils.version import LooseVersion

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.remote_server import RemoteServer


class DvtSecureSocketProxyService(RemoteServer):
    SERVICE_NAME = 'com.apple.instruments.remoteserver.DVTSecureSocketProxy'
    OLD_SERVICE_NAME = 'com.apple.instruments.remoteserver'

    def __init__(self, lockdown: LockdownClient):
        if LooseVersion(lockdown.ios_version) >= LooseVersion('14.0'):
            service_name = self.SERVICE_NAME
            remove_ssl_context = False
        else:
            service_name = self.OLD_SERVICE_NAME
            remove_ssl_context = True

        super().__init__(lockdown, service_name, remove_ssl_context=remove_ssl_context)



import zipfile
from contextlib import contextmanager
from io import BytesIO

from cached_property import cached_property
from pymobiledevice3.restore.ipsw.build_manifest import BuildManifest
from pymobiledevice3.restore.ipsw.firmware import Firmware


class IPSW:
    def __init__(self, file: BytesIO):
        self._archive = zipfile.ZipFile(file)
        self.build_manifest = BuildManifest(self, self._archive.read('BuildManifest.plist'))

    @cached_property
    def restore_version(self):
        return self.read('RestoreVersion.plist')

    @cached_property
    def system_version(self):
        return self.read('SystemVersion.plist')

    @contextmanager
    def open_path(self, path: str):
        file = self._archive.open(path)
        try:
            yield file
        finally:
            file.close()

    def read(self, path: str):
        return self._archive.read(path)

    def get_global_manifest(self, macos_variant: str, device_class: str):
        manifest_path = f'Firmware/Manifests/restore/{macos_variant}/apticket.{device_class}.im4m'
        return self.read(manifest_path)

    def get_firmware(self, firmware_path: str):
        return Firmware(firmware_path, self)



#!/usr/bin/env python3
import logging
import plistlib
import shlex
from contextlib import closing
from pathlib import Path
import time

import requests
import tqdm

from pymobiledevice3.lockdown import LockdownClient

ACTIVATION_USER_AGENT_IOS = 'iOS Device Activator (MobileActivation-20 built on Jan 15 2012 at 19:07:28)'
ACTIVATION_DEFAULT_URL = 'https://albert.apple.com/deviceservices/deviceActivation'
ACTIVATION_DRM_HANDSHAKE_DEFAULT_URL = 'https://albert.apple.com/deviceservices/drmHandshake'
DEFAULT_HEADERS = {
    'Accept': 'application/xml',
    'User-Agent': ACTIVATION_USER_AGENT_IOS,
    'Expect': '100-continue',
}

ACTIVATION_REQUESTS_SUBDIR = Path('offline_requests')
NONCE_CYCLE_INTERVAL = 60 * 5


class MobileActivationService:
    SERVICE_NAME = 'com.apple.mobileactivationd'

    def __init__(self, lockdown: LockdownClient, offline=True, now=False):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.offline = offline
        self._now = now
        self._offline_start = 0
        self._offline_end = 0

    @property
    def state(self):
        return self.send_command('GetActivationStateRequest')['Value']

    def activate(self):
        blob = self.create_activation_session_info()

        if self.offline and not self._now:
            self.logger.info('waiting for the next 5 minutes cycle')

            handshake_request_message = blob['HandshakeRequestMessage']
            while handshake_request_message == blob['HandshakeRequestMessage']:
                blob = self.create_activation_session_info()

            self.logger.info('new cycle started! go go go!')
            self._offline_start = time.time()
            self._offline_end = self._offline_start + NONCE_CYCLE_INTERVAL

        # create drmHandshake request with blob from device
        headers = {'Content-Type': 'application/x-apple-plist'}
        headers.update(DEFAULT_HEADERS)
        content, headers = self.post(ACTIVATION_DRM_HANDSHAKE_DEFAULT_URL, data=plistlib.dumps(blob), headers=headers)

        activation_info = self.create_activation_info_with_session(content)

        content, headers = self.post(ACTIVATION_DEFAULT_URL, data={'activation-info': plistlib.dumps(activation_info)})
        assert headers['Content-Type'] == 'text/xml'
        self.activate_with_session(content, headers)

    def deactivate(self):
        return self.send_command('DeactivateRequest')

    def create_activation_session_info(self):
        return self.send_command('CreateTunnel1SessionInfoRequest')['Value']

    def create_activation_info_with_session(self, handshake_response):
        return self.send_command('CreateTunnel1ActivationInfoRequest', handshake_response)['Value']

    def activate_with_session(self, activation_record, headers):
        data = {
            'Command': 'HandleActivationInfoWithSessionRequest',
            'Value': activation_record,
        }
        if headers:
            data['ActivationResponseHeaders'] = dict(headers)
        with closing(LockdownClient(self.lockdown.udid).start_service(self.SERVICE_NAME)) as service:
            return service.send_recv_plist(data)

    def send_command(self, command, value=''):
        data = {'Command': command}
        if value:
            data['Value'] = value
        with closing(LockdownClient(self.lockdown.udid).start_service(self.SERVICE_NAME)) as service:
            return service.send_recv_plist(data)

    def post(self, url, data, headers=None):
        if headers is None:
            headers = DEFAULT_HEADERS
        if not self.offline:
            resp = requests.post(url, data=data, headers=headers)
            return resp.content, resp.headers
        ACTIVATION_REQUESTS_SUBDIR.mkdir(parents=True, exist_ok=True)
        for file in ACTIVATION_REQUESTS_SUBDIR.iterdir():
            file.unlink()
        curl_arguments = ['curl', '-X', 'POST', '-D', 'headers.txt']

        if isinstance(data, bytes):
            request_data = ACTIVATION_REQUESTS_SUBDIR / f'request_data_{hash(data)}.plist'
            request_data.write_bytes(data)
            curl_arguments.extend(['--data-binary', f'@{request_data.name}'])
        elif isinstance(data, dict):
            for key, value, in data.items():
                request_data = ACTIVATION_REQUESTS_SUBDIR / f'request_data_{hash(value)}.plist'
                request_data.write_bytes(value)
                curl_arguments.extend(['-F', f'{key}=@{request_data.name}'])

        for header, value in headers.items():
            curl_arguments.extend(['-H', f'{header}: {value}'])

        headers = ACTIVATION_REQUESTS_SUBDIR / 'headers.txt'
        curl_arguments.extend(['-D', headers.name])
        curl_arguments.append(url)

        response = ACTIVATION_REQUESTS_SUBDIR / f'request_data_{time.time()}.txt'
        request = ACTIVATION_REQUESTS_SUBDIR / f'request_script.sh'
        request.write_text(f'#!/bin/sh\n{shlex.join(curl_arguments)} | tee {response.name}\n')
        request.chmod(0o755)

        self.logger.info(f'Run the following shell script ({request.name})')

        # Check for plist response.
        if not self._now:
            pb = tqdm.tqdm(total=int(self._offline_end - self._offline_start), desc='Time Left', dynamic_ncols=True)
            pb.update(int(time.time() - self._offline_start))
            while not response.exists() or b'</plist>' not in response.read_bytes():
                time.sleep(1)
                pb.update(1)
            pb.close()

        # Check for headers.
        while not headers.exists() or ': ' not in headers.read_text():
            time.sleep(1)

        raw_headers = headers.read_text()
        parsed_headers = {}
        for line in raw_headers.splitlines():
            if ': ' not in line:
                continue
            key, value = line.split(': ', 1)
            parsed_headers[key] = value

        return response.read_bytes(), parsed_headers



from io import BytesIO

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient


class MisagentService(object):
    SERVICE_NAME = 'com.apple.misagent'

    def __init__(self, lockdown: LockdownClient):
        self.lockdown = lockdown
        self.service = self.lockdown.start_service(self.SERVICE_NAME)

    def install(self, plist: BytesIO):
        response = self.service.send_recv_plist({'MessageType': 'Install',
                                                 'Profile': plist.read(),
                                                 'ProfileType': 'Provisioning'})
        if response['Status']:
            raise PyMobileDevice3Exception(f'invalid status: {response}')

        return response

    def remove(self, profile_id):
        response = self.service.send_recv_plist({'MessageType': 'Remove',
                                                 'ProfileID': profile_id,
                                                 'ProfileType': 'Provisioning'})
        if response['Status']:
            raise PyMobileDevice3Exception(f'invalid status: {response}')

        return response

    def copy_all(self):
        response = self.service.send_recv_plist({'MessageType': 'CopyAll',
                                                 'ProfileType': 'Provisioning'})
        if response['Status']:
            raise PyMobileDevice3Exception(f'invalid status: {response}')

        return response['Payload']



import click

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.services.companion import CompanionProxyService


@click.group()
def cli():
    """ companion cli """
    pass


@cli.group()
def companion():
    """ companion options """
    pass


@companion.command('list', cls=Command)
@click.option('--color/--no-color', default=True)
def companion_list(lockdown, color):
    """ list all paired companion devices """
    print_json(CompanionProxyService(lockdown).list(), colored=color, default=lambda x: '<non-serializable>')



import logging
import plistlib
import time
from pathlib import Path
from uuid import uuid4

import asn1
import requests

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.restore.img4 import img4_get_component_tag
from pymobiledevice3.utils import bytes_to_uint

TSS_CLIENT_VERSION_STRING = 'libauthinstall-776.60.1'
TICKETS_SUBDIR = Path('offline_requests')

OFFLINE_REQUEST_SCRIPT = """#!/bin/sh
curl -d "@{request}" -H 'Cache-Control: no-cache' -H 'Content-type: text/xml; charset="utf-8"' -H 'User-Agent: InetURL/1.0' -H 'Expect: ' 'http://gs.apple.com/TSS/controller?action=2' | tee {response} 
"""


class TSSResponse(dict):
    @property
    def ap_img4_ticket(self):
        ticket = self.get('ApImg4Ticket')

        if ticket is None:
            raise PyMobileDevice3Exception('TSS response doesn\'t contain a ApImg4Ticket')

        return ticket

    @property
    def bb_ticket(self):
        return self.get('BBTicket')

    def get_path_by_entry(self, component: str):
        node = self.get(component)
        if node is not None:
            return node.get('Path')

        return None


class TSSRequest:
    def __init__(self, offline=False):
        self._request = {
            '@BBTicket': True,
            '@HostPlatformInfo': 'mac',
            '@VersionInfo': TSS_CLIENT_VERSION_STRING,
            '@UUID': str(uuid4()).upper(),
        }
        self._offline = offline

    @staticmethod
    def apply_restore_request_rules(tss_entry: dict, parameters: dict, rules: list):
        for rule in rules:
            conditions_fulfilled = True
            conditions = rule['Conditions']
            for key, value in conditions.items():
                if not conditions_fulfilled:
                    break

                if key == 'ApRawProductionMode':
                    value2 = parameters.get('ApProductionMode')
                elif key == 'ApCurrentProductionMode':
                    value2 = parameters.get('ApProductionMode')
                elif key == 'ApRawSecurityMode':
                    value2 = parameters.get('ApSecurityMode')
                elif key == 'ApRequiresImage4':
                    value2 = parameters.get('ApSupportsImg4')
                elif key == 'ApDemotionPolicyOverride':
                    value2 = parameters.get('DemotionPolicy')
                elif key == 'ApInRomDFU':
                    value2 = parameters.get('ApInRomDFU')
                else:
                    logging.error(f'Unhandled condition {key} while parsing RestoreRequestRules')
                    value2 = None

                if value2:
                    conditions_fulfilled = value == value2
                else:
                    conditions_fulfilled = False

            if not conditions_fulfilled:
                continue

            actions = rule['Actions']
            for key, value in actions.items():
                if value != 255:
                    value2 = tss_entry.get(key)
                    if value2:
                        tss_entry.pop(key)
                    logging.debug(f'Adding {key}={value} to TSS entry')
                    tss_entry[key] = value
        return tss_entry

    def add_common_tags(self, parameters: dict, overrides=None):
        keys = ('ApECID', 'UniqueBuildID', 'ApChipID', 'ApBoardID', 'ApSecurityDomain')
        for k in keys:
            if k in parameters:
                self._request[k] = parameters[k]
        if overrides is not None:
            self._request.update(overrides)

    def add_local_policy_tags(self, parameters: dict):
        self._request['@ApImg4Ticket'] = True

        keys_to_copy = (
            'Ap,LocalBoot', 'Ap,LocalPolicy', 'Ap,NextStageIM4MHash', 'Ap,NextStageIM4MHash',
            'Ap,RecoveryOSPolicyNonceHash', 'Ap,VolumeUUID', 'ApECID', 'ApChipID', 'ApBoardID', 'ApSecurityDomain'
                                                                                                'ApNonce',
            'ApSecurityMode', 'ApProductionMode')

        for k in keys_to_copy:
            if k in parameters:
                self._request[k] = parameters[k]

    def add_vinyl_tags(self, parameters: dict, overrides=None):
        self._request['@BBTicket'] = True
        self._request['@eUICC,Ticket'] = True

        keys = ('eUICC,ChipID', 'eUICC,EID', 'eUICC,RootKeyIdentifier')
        for k in keys:
            if k in parameters:
                self._request[k] = parameters[k]

        # set Nonce for eUICC,Gold component
        node = parameters.get('EUICCGoldNonce')
        if node is not None:
            n = self._request.get('eUICC,Gold')
            if n is not None:
                n['Nonce'] = node

        # set Nonce for eUICC,Main component
        node = parameters.get('EUICCMainNonce')
        if node is not None:
            n = self._request.get('eUICC,Main')
            if n is not None:
                n['Nonce'] = node

        if overrides is not None:
            self._request.update(overrides)

    def add_ap_tags(self, parameters: dict, overrides=None):
        """ loop over components from build manifest """

        manifest_node = parameters['Manifest']

        # add components to request
        skipped_keys = ('BasebandFirmware', 'SE,UpdatePayload', 'BaseSystem', 'Diags',)
        for key, manifest_entry in manifest_node.items():
            if key in skipped_keys:
                continue

            info_dict = manifest_entry.get('Info')
            if info_dict is None:
                continue

            if parameters.get('ApSupportsImg4', False) and ('RestoreRequestRules' not in info_dict):
                logging.debug(f'Skipping "{key}" as it doesn\'t have RestoreRequestRules')
                continue

            if parameters.get('_OnlyFWComponents', False):
                if not manifest_node.get('Trusted', False):
                    logging.debug(f'skipping {key} as it is not trusted')
                    continue
                info = manifest_node['Info']
                if not info['IsFirmwarePayload'] and not info['IsSecondaryFirmwarePayload'] and \
                        not info['IsFUDFirmware']:
                    logging.debug(f'skipping {key} as it is neither firmware nor secondary nor FUD firmware payload')
                    continue

            # copy this entry
            tss_entry = dict(manifest_entry)

            # remove obsolete Info node
            tss_entry.pop('Info')

            # handle RestoreRequestRules
            if 'Info' in manifest_entry and 'RestoreRequestRules' in manifest_entry['Info']:
                rules = manifest_entry['Info']['RestoreRequestRules']
                if rules:
                    logging.debug(f'Applying restore request rules for entry {key}')
                    tss_entry = self.apply_restore_request_rules(tss_entry, parameters, rules)

            # Make sure we have a Digest key for Trusted items even if empty
            node = manifest_entry.get('Trusted', False)
            if node:
                if manifest_entry.get('Digest') is None:
                    tss_entry['Digest'] = bytes()

            self._request[key] = tss_entry

        if overrides is not None:
            self._request.update(overrides)

    def add_ap_img3_tags(self, parameters: dict):
        if 'ApNonce' in parameters:
            self._request['ApNonce'] = parameters['ApNonce']
        self._request['@APTicket'] = True

    def add_ap_img4_tags(self, parameters):
        keys_to_copy = (
            'ApNonce', 'Ap,OSLongVersion', 'ApSecurityMode', 'ApProductionMode', 'ApSepNonce',
            'PearlCertificationRootPub')
        for k in keys_to_copy:
            if k in parameters:
                v = parameters[k]
                if k == 'ApSepNonce':
                    k = 'SepNonce'
                self._request[k] = v

        self._request['@ApImg4Ticket'] = True

    def add_se_tags(self, parameters: dict, overrides=None):
        manifest = parameters['Manifest']

        # add tags indicating we want to get the SE,Ticket
        self._request['@BBTicket'] = True
        self._request['@SE,Ticket'] = True

        keys_to_copy = ('SE,ChipID', 'SE,ID', 'SE,Nonce', 'SE,Nonce', 'SE,RootKeyIdentifier',
                        'SEChipID', 'SEID', 'SENonce', 'SENonce', 'SERootKeyIdentifier',)

        for src_key in keys_to_copy:
            if src_key not in parameters:
                continue

            if src_key.startswith('SE'):
                dst_key = src_key
                if not dst_key.startswith('SE,'):
                    # make sure there is a comma (,) after prefix
                    dst_key = 'SE,' + dst_key.split('SE', 1)[1]
                self._request[dst_key] = parameters[src_key]

        # 'IsDev' determines whether we have Production or Development
        is_dev = parameters.get('SE,IsDev')
        if is_dev is None:
            is_dev = parameters.get('SEIsDev', False)

        # add SE,* components from build manifest to request
        for key, manifest_entry in manifest.items():
            if not key.startswith('SE'):
                continue

            # copy this entry
            tss_entry = dict(manifest_entry)

            # remove Info node
            tss_entry.pop('Info')

            # remove Development or Production key/hash node
            if is_dev:
                if 'ProductionCMAC' in tss_entry:
                    tss_entry.pop('ProductionCMAC')
                if 'ProductionUpdatePayloadHash' in tss_entry:
                    tss_entry.pop('ProductionUpdatePayloadHash')
            else:
                if 'DevelopmentCMAC' in tss_entry:
                    tss_entry.pop('DevelopmentCMAC')
                if 'DevelopmentUpdatePayloadHash' in tss_entry:
                    tss_entry.pop('DevelopmentUpdatePayloadHash')

            # add entry to request
            self._request[key] = tss_entry

        if overrides is not None:
            self._request.update(overrides)

    def add_savage_tags(self, parameters: dict, overrides=None, component_name=None):
        manifest = parameters['Manifest']

        # add tags indicating we want to get the Savage,Ticket
        self._request['@BBTicket'] = True
        self._request['@Savage,Ticket'] = True

        # add Savage,UID
        self._request['Savage,UID'] = parameters['Savage,UID']

        # add SEP
        self._request['SEP'] = {'Digest': manifest['SEP']['Digest']}

        keys_to_copy = (
            'Savage,PatchEpoch', 'Savage,ChipID', 'Savage,AllowOfflineBoot', 'Savage,ReadFWKey',
            'Savage,ProductionMode', 'Savage,Nonce', 'Savage,Nonce')

        for k in keys_to_copy:
            if k in parameters:
                self._request[k] = parameters[k]

        isprod = parameters['Savage,ProductionMode']

        # get the right component name
        comp_name = 'Savage,B0-Prod-Patch' if isprod else 'Savage,B0-Dev-Patch'
        node = parameters.get('Savage,Revision')

        if isinstance(node, bytes):
            savage_rev = node
            if ((savage_rev[0] | 0x10) & 0xF0) == 0x30:
                comp_name = 'Savage,B2-Prod-Patch' if isprod else 'Savage,B2-Dev-Patch'
            elif (savage_rev[0] & 0xF0) == 0xA0:
                comp_name = 'Savage,BA-Prod-Patch' if isprod else 'Savage,BA-Dev-Patch'

        # add Savage,B?-*-Patch
        d = dict(manifest[comp_name])
        d.pop('Info')
        self._request[comp_name] = d

        if overrides is not None:
            self._request.update(overrides)

        return comp_name

    def add_yonkers_tags(self, parameters: dict, overrides=None):
        manifest = parameters['Manifest']

        # add tags indicating we want to get the Yonkers,Ticket
        self._request['@BBTicket'] = True
        self._request['@Yonkers,Ticket'] = True

        # add SEP
        self._request['SEP'] = {'Digest': manifest['SEP']['Digest']}

        keys_to_copy = (
            'Yonkers,AllowOfflineBoot', 'Yonkers,BoardID', 'Yonkers,ChipID', 'Yonkers,ECID', 'Yonkers,Nonce',
            'Yonkers,PatchEpoch', 'Yonkers,ProductionMode', 'Yonkers,ReadECKey', 'Yonkers,ReadFWKey',)

        for k in keys_to_copy:
            if k in parameters:
                self._request[k] = parameters[k]

        isprod = parameters.get('Yonkers,ProductionMode', 1)
        fabrevision = parameters.get('Yonkers,FabRevision', 0xffffffffffffffff)
        comp_node = None
        result_comp_name = None

        for comp_name, node in manifest.items():
            if not comp_name.startswith('Yonkers,'):
                continue

            target_node = 1
            sub_node = node.get('EPRO')
            if sub_node:
                target_node &= sub_node if isprod else not sub_node
            sub_node = node.get('FabRevision')
            if sub_node:
                target_node &= sub_node == fabrevision

            if target_node:
                comp_node = node
                result_comp_name = comp_name
                break

        if comp_node is None:
            raise PyMobileDevice3Exception(f'No Yonkers node for {isprod}/{fabrevision}')

        # add Yonkers,SysTopPatch
        comp_dict = dict(comp_node)
        comp_dict.pop('Info')
        self._request[result_comp_name] = comp_dict

        if overrides is not None:
            self._request.update(overrides)

        return result_comp_name

    def add_baseband_tags(self, parameters: dict, overrides=None):
        self._request['@BBTicket'] = True

        keys_to_copy = (
            'BbChipID', 'BbProvisioningManifestKeyHash', 'BbActivationManifestKeyHash', 'BbCalibrationManifestKeyHash',
            'BbFactoryActivationManifestKeyHash', 'BbFDRSecurityKeyHash', 'BbSkeyId', 'BbNonce',
            'BbGoldCertId', 'BbSNUM',)

        for k in keys_to_copy:
            if k in parameters:
                self._request[k] = parameters[k]

        bb_chip_id = parameters['BbChipID']
        bb_cert_id = parameters['BbGoldCertId']

        bbfwdict = dict(parameters['Manifest']['BasebandFirmware'])
        bbfwdict.pop('Info')

        if bb_chip_id == 0x68:
            # depending on the BasebandCertId remove certain nodes
            if bb_cert_id in (0x26F3FACC, 0x5CF2EC4E, 0x8399785A):
                bbfwdict.pop('PSI2-PartialDigest')
                bbfwdict.pop('RestorePSI2-PartialDigest')
            else:
                bbfwdict.pop('PSI-PartialDigest')
                bbfwdict.pop('RestorePSI-PartialDigest')

        self._request['BasebandFirmware'] = bbfwdict

        if overrides:
            self._request.update(overrides)

    def add_rose_tags(self, parameters: dict, overrides: dict = None):
        manifest = parameters['Manifest']

        # add tags indicating we want to get the Rap,Ticket
        self._request['@BBTicket'] = True
        self._request['@Rap,Ticket'] = True

        keys_to_copy_uint = ('Rap,BoardID', 'Rap,ChipID', 'Rap,ECID', 'Rap,SecurityDomain',)

        for k in keys_to_copy_uint:
            self._request[k] = bytes_to_uint(parameters[k])

        keys_to_copy_bool = ('Rap,ProductionMode', 'Rap,SecurityMode',)

        for k in keys_to_copy_bool:
            self._request[k] = bytes_to_uint(parameters[k]) == 1

        nonce = parameters.get('Rap,Nonce')
        if nonce is not None:
            self._request['Rap,Nonce'] = nonce

        for comp_name, node in manifest.items():
            if not comp_name.startswith('Rap,'):
                continue

            manifest_entry = dict(node)

            # handle RestoreRequestRules
            rules = manifest_entry['Info'].get('RestoreRequestRules')
            if rules is not None:
                self.apply_restore_request_rules(manifest_entry, parameters, rules)

            # Make sure we have a Digest key for Trusted items even if empty
            trusted = manifest_entry.get('Trusted', False)

            if trusted:
                digest = manifest_entry.get('Digest')
                if digest is None:
                    logging.debug(f'No Digest data, using empty value for entry {comp_name}')
                    manifest_entry['Digest'] = b''

            manifest_entry.pop('Info')

            # finally add entry to request
            self._request[comp_name] = manifest_entry

        if overrides is not None:
            self._request.update(overrides)

    def add_veridian_tags(self, parameters: dict, overrides: dict = None):
        manifest = parameters['Manifest']

        # add tags indicating we want to get the Rap,Ticket
        self._request['@BBTicket'] = True
        self._request['@BMU,Ticket'] = True

        self._request['BMU,ChipID'] = parameters['ChipID']
        self._request['BMU,UniqueID'] = parameters['UniqueID']
        self._request['BMU,ProductionMode'] = parameters['ProductionMode']

        nonce = parameters.get('Nonce')
        if nonce is not None:
            self._request['BMU,Nonce'] = nonce

        for comp_name, node in manifest.items():
            if not comp_name.startswith('BMU,'):
                continue

            manifest_entry = dict(node)

            # handle RestoreRequestRules
            rules = manifest_entry['Info'].get('RestoreRequestRules')
            if rules is not None:
                self.apply_restore_request_rules(manifest_entry, parameters, rules)

            # Make sure we have a Digest key for Trusted items even if empty
            trusted = manifest_entry.get('Trusted', False)

            if trusted:
                digest = manifest_entry.get('Digest')
                if digest is None:
                    logging.debug(f'No Digest data, using empty value for entry {comp_name}')
                    manifest_entry['Digest'] = b''

            manifest_entry.pop('Info')

            # finally add entry to request
            self._request[comp_name] = manifest_entry

        if overrides is not None:
            self._request.update(overrides)

    def img4_create_local_manifest(self, build_identity=None):
        manifest = None
        if build_identity is not None:
            manifest = build_identity['Manifest']

        p = asn1.Encoder()
        p.start()

        p.write(b'MANP', asn1.Numbers.IA5String)
        p.enter(asn1.Numbers.Set)

        p.write(b'BORD', asn1.Numbers.IA5String)
        p.write(self._request['ApBoardID'], asn1.Numbers.Integer)

        p.write(b'CEPO', asn1.Numbers.IA5String)
        p.write(0, asn1.Numbers.Integer)

        p.write(b'CHIP', asn1.Numbers.IA5String)
        p.write(self._request['ApChipID'], asn1.Numbers.Integer)

        p.write(b'CPRO', asn1.Numbers.IA5String)
        p.write(self._request['ApProductionMode'], asn1.Numbers.Integer)

        p.write(b'CSEC', asn1.Numbers.IA5String)
        p.write(0, asn1.Numbers.Integer)

        p.write(b'SDOM', asn1.Numbers.IA5String)
        p.write(self._request['ApSecurityDomain'], asn1.Numbers.Integer)

        p.leave()

        # now write the components
        for k, v in self._request.items():
            if isinstance(v, dict):
                # check if component has Img4PayloadType
                comp = None
                if manifest is not None:
                    comp = manifest[k]['Info'].get('Img4PayloadType')

                if comp is None:
                    comp = img4_get_component_tag(k)

                if comp is None:
                    raise NotImplementedError(f'Unhandled component {k} - can\'t create manifest')

                logging.debug(f'found component {comp} ({k})')

        # write manifest body header
        p.write(b'MANB', asn1.Numbers.IA5String)
        p.enter(asn1.Numbers.Set)
        p.leave()

        # write header values
        p.write(b'IM4M', asn1.Numbers.IA5String)
        p.write(0, asn1.Numbers.Integer)

        return p.output()

    def update(self, options):
        self._request.update(options)

    def send_receive(self) -> TSSResponse:
        headers = {
            'Cache-Control': 'no-cache',
            'Content-type': 'text/xml; charset="utf-8"',
            'User-Agent': 'InetURL/1.0',
            'Expect': '',
        }

        if self._offline:
            unique_identifier = str(time.time())
            request_plist_path = TICKETS_SUBDIR / f'request_data_{unique_identifier}.plist'
            request_script_path = TICKETS_SUBDIR / f'request_script.sh'
            response_path = TICKETS_SUBDIR / f'response_{unique_identifier}.txt'

            TICKETS_SUBDIR.mkdir(parents=True, exist_ok=True)
            for file in TICKETS_SUBDIR.iterdir():
                file.unlink()

            with request_plist_path.open('wb') as f:
                plistlib.dump(self._request, f)

            request_script_path.write_text(
                OFFLINE_REQUEST_SCRIPT.format(request=request_plist_path.name, response=response_path.name))
            request_script_path.chmod(0o755)

            logging.info(f'waiting for {response_path} to be created')
            while not response_path.exists() or b'</plist>' not in response_path.read_bytes():
                time.sleep(1)

            content = response_path.read_bytes()
        else:
            logging.info(f'Sending TSS request...')
            r = requests.post('http://gs.apple.com/TSS/controller?action=2', headers=headers,
                              data=plistlib.dumps(self._request), verify=False)
            content = r.content

        if b'MESSAGE=SUCCESS' in content:
            logging.info('response successfully received')

        message = content.split(b'MESSAGE=', 1)[1].split(b'&', 1)[0].decode()
        if message != 'SUCCESS':
            raise Exception(f'server replied: {message}')

        return TSSResponse(plistlib.loads(content.split(b'REQUEST_STRING=', 1)[1]))



import pytest

from pymobiledevice3.lockdown import LockdownClient


@pytest.fixture(scope='function')
def lockdown():
    """
    Creates a new lockdown client for each test.
    """
    return LockdownClient()



import logging

import click

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.exceptions import NotMountedError, UnsupportedCommandError, AlreadyMountedError
from pymobiledevice3.services.mobile_image_mounter import MobileImageMounterService


@click.group()
def cli():
    """ mounter cli """
    pass


@cli.group()
def mounter():
    """ mounter options """
    pass


@mounter.command('list', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_list(lockdown, color):
    """ list all mounted images """
    output = []

    images = MobileImageMounterService(lockdown=lockdown).list_images()['EntryList']
    for image in images:
        image['ImageSignature'] = image['ImageSignature'].hex()
        output.append(image)

    print_json(output, colored=color)


@mounter.command('lookup', cls=Command)
@click.option('--color/--no-color', default=True)
@click.argument('image_type')
def mounter_lookup(lockdown, color, image_type):
    """ lookup mounter image type """
    output = []

    signatures = MobileImageMounterService(lockdown=lockdown).lookup_image(image_type)['ImageSignature']
    for signature in signatures:
        output.append(signature.hex())

    print_json(output, colored=color)


@mounter.command('umount', cls=Command)
def mounter_umount(lockdown):
    """ unmount developer image. """
    image_type = 'Developer'
    mount_path = '/Developer'
    image_mounter = MobileImageMounterService(lockdown=lockdown)
    try:
        image_mounter.umount(image_type, mount_path, b'')
        logging.info('DeveloperDiskImage umounted successfully')
    except NotMountedError:
        logging.error('DeveloperDiskImage isn\'t currently mounted')
    except UnsupportedCommandError:
        logging.error('Your iOS version doesn\'t support this command')


@mounter.command('mount', cls=Command)
@click.option('-i', '--image', type=click.Path(exists=True))
@click.option('-s', '--signature', type=click.Path(exists=True))
@click.option('-x', '--xcode', type=click.Path(exists=True, dir_okay=True, file_okay=False),
              default='/Applications/Xcode.app',
              help='Xcode application path used to figure out automatically the DeveloperDiskImage path')
@click.option('-v', '--version', help='use a different DeveloperDiskImage version from the one retrieved by lockdown'
                                      'connection')
def mounter_mount(lockdown, image, signature, xcode, version):
    """ mount developer image. """
    image_type = 'Developer'

    if image and signature:
        logging.debug('using given image and signature for mount command')
    else:
        logging.debug('trying to figure out the best suited DeveloperDiskImage')
        if version is None:
            version = lockdown.sanitized_ios_version
        image = f'{xcode}/Contents/Developer/Platforms/iPhoneOS.platform/DeviceSupport/{version}/DeveloperDiskImage.dmg'
        signature = f'{image}.signature'

    with open(image, 'rb') as image:
        image = image.read()

    with open(signature, 'rb') as signature:
        signature = signature.read()

    image_mounter = MobileImageMounterService(lockdown=lockdown)
    image_mounter.upload_image(image_type, image, signature)
    try:
        image_mounter.mount(image_type, signature)
        logging.info('DeveloperDiskImage mounted successfully')
    except AlreadyMountedError:
        logging.error('DeveloperDiskImage is already mounted')



#!/usr/bin/env python3
import logging
import time

from pymobiledevice3.lockdown import LockdownClient


class HeartbeatService(object):
    SERVICE_NAME = 'com.apple.mobile.heartbeat'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown

    def start(self, interval=None):
        start = time.time()
        service = self.lockdown.start_service(self.SERVICE_NAME)

        while True:
            response = service.recv_plist()
            self.logger.debug(response)

            if interval is not None:
                if time.time() >= start + interval:
                    break

            service.send_plist({'Command': 'Polo'})



import pytest

from pymobiledevice3.utils import sanitize_ios_version


@pytest.mark.parametrize('version, sanitized', [
    ('14.5', '14.5'),
    ('14.5.1', '14.5'),
    ('0.0', '0.0'),
    ('9999.9999', '9999.9999'),
    ('9999.9999.9999', '9999.9999'),
])
def test_sanitize_ios_version(version, sanitized):
    assert sanitize_ios_version(version) == sanitized



from collections import namedtuple

IRecvDevice = namedtuple('IRecvDevice', 'product_type hardware_model board_id chip_id display_name')

IRECV_DEVICES = (
    # iPhone
    IRecvDevice(product_type='iPhone1,1', hardware_model='m68ap', board_id=0x00, chip_id=0x8900,
                display_name='iPhone 2G'),
    IRecvDevice(product_type='iPhone1,2', hardware_model='n82ap', board_id=0x04, chip_id=0x8900,
                display_name='iPhone 3G'),
    IRecvDevice(product_type='iPhone2,1', hardware_model='n88ap', board_id=0x00, chip_id=0x8920,
                display_name='iPhone 3Gs'),
    IRecvDevice(product_type='iPhone3,1', hardware_model='n90ap', board_id=0x00, chip_id=0x8930,
                display_name='iPhone 4 (GSM)'),
    IRecvDevice(product_type='iPhone3,2', hardware_model='n90bap', board_id=0x04, chip_id=0x8930,
                display_name='iPhone 4 (GSM) R2 2012'),
    IRecvDevice(product_type='iPhone3,3', hardware_model='n92ap', board_id=0x06, chip_id=0x8930,
                display_name='iPhone 4 (CDMA)'),
    IRecvDevice(product_type='iPhone4,1', hardware_model='n94ap', board_id=0x08, chip_id=0x8940,
                display_name='iPhone 4s'),
    IRecvDevice(product_type='iPhone5,1', hardware_model='n41ap', board_id=0x00, chip_id=0x8950,
                display_name='iPhone 5 (GSM)'),
    IRecvDevice(product_type='iPhone5,2', hardware_model='n42ap', board_id=0x02, chip_id=0x8950,
                display_name='iPhone 5 (Global)'),
    IRecvDevice(product_type='iPhone5,3', hardware_model='n48ap', board_id=0x0a, chip_id=0x8950,
                display_name='iPhone 5c (GSM)'),
    IRecvDevice(product_type='iPhone5,4', hardware_model='n49ap', board_id=0x0e, chip_id=0x8950,
                display_name='iPhone 5c (Global)'),
    IRecvDevice(product_type='iPhone6,1', hardware_model='n51ap', board_id=0x00, chip_id=0x8960,
                display_name='iPhone 5s (GSM)'),
    IRecvDevice(product_type='iPhone6,2', hardware_model='n53ap', board_id=0x02, chip_id=0x8960,
                display_name='iPhone 5s (Global)'),
    IRecvDevice(product_type='iPhone7,1', hardware_model='n56ap', board_id=0x04, chip_id=0x7000,
                display_name='iPhone 6 Plus'),
    IRecvDevice(product_type='iPhone7,2', hardware_model='n61ap', board_id=0x06, chip_id=0x7000,
                display_name='iPhone 6'),
    IRecvDevice(product_type='iPhone8,1', hardware_model='n71ap', board_id=0x04, chip_id=0x8000,
                display_name='iPhone 6s'),
    IRecvDevice(product_type='iPhone8,1', hardware_model='n71map', board_id=0x04, chip_id=0x8003,
                display_name='iPhone 6s'),
    IRecvDevice(product_type='iPhone8,2', hardware_model='n66ap', board_id=0x06, chip_id=0x8000,
                display_name='iPhone 6s Plus'),
    IRecvDevice(product_type='iPhone8,2', hardware_model='n66map', board_id=0x06, chip_id=0x8003,
                display_name='iPhone 6s Plus'),
    IRecvDevice(product_type='iPhone8,4', hardware_model='n69ap', board_id=0x02, chip_id=0x8003,
                display_name='iPhone SE'),
    IRecvDevice(product_type='iPhone8,4', hardware_model='n69uap', board_id=0x02, chip_id=0x8000,
                display_name='iPhone SE'),
    IRecvDevice(product_type='iPhone9,1', hardware_model='d10ap', board_id=0x08, chip_id=0x8010,
                display_name='iPhone 7 (Global)'),
    IRecvDevice(product_type='iPhone9,2', hardware_model='d11ap', board_id=0x0a, chip_id=0x8010,
                display_name='iPhone 7 Plus (Global)'),
    IRecvDevice(product_type='iPhone9,3', hardware_model='d101ap', board_id=0x0c, chip_id=0x8010,
                display_name='iPhone 7 (GSM)'),
    IRecvDevice(product_type='iPhone9,4', hardware_model='d111ap', board_id=0x0e, chip_id=0x8010,
                display_name='iPhone 7 Plus (GSM)'),
    IRecvDevice(product_type='iPhone10,1', hardware_model='d20ap', board_id=0x02, chip_id=0x8015,
                display_name='iPhone 8 (Global)'),
    IRecvDevice(product_type='iPhone10,2', hardware_model='d21ap', board_id=0x04, chip_id=0x8015,
                display_name='iPhone 8 Plus (Global)'),
    IRecvDevice(product_type='iPhone10,3', hardware_model='d22ap', board_id=0x06, chip_id=0x8015,
                display_name='iPhone X (Global)'),
    IRecvDevice(product_type='iPhone10,4', hardware_model='d201ap', board_id=0x0a, chip_id=0x8015,
                display_name='iPhone 8 (GSM)'),
    IRecvDevice(product_type='iPhone10,5', hardware_model='d211ap', board_id=0x0c, chip_id=0x8015,
                display_name='iPhone 8 Plus (GSM)'),
    IRecvDevice(product_type='iPhone10,6', hardware_model='d221ap', board_id=0x0e, chip_id=0x8015,
                display_name='iPhone X (GSM)'),
    IRecvDevice(product_type='iPhone11,2', hardware_model='d321ap', board_id=0x0e, chip_id=0x8020,
                display_name='iPhone XS'),
    IRecvDevice(product_type='iPhone11,4', hardware_model='d331ap', board_id=0x0a, chip_id=0x8020,
                display_name='iPhone XS Max (China)'),
    IRecvDevice(product_type='iPhone11,6', hardware_model='d331pap', board_id=0x1a, chip_id=0x8020,
                display_name='iPhone XS Max'),
    IRecvDevice(product_type='iPhone11,8', hardware_model='n841ap', board_id=0x0c, chip_id=0x8020,
                display_name='iPhone XR'),
    IRecvDevice(product_type='iPhone12,1', hardware_model='n104ap', board_id=0x04, chip_id=0x8030,
                display_name='iPhone 11'),
    IRecvDevice(product_type='iPhone12,3', hardware_model='d421ap', board_id=0x06, chip_id=0x8030,
                display_name='iPhone 11 Pro'),
    IRecvDevice(product_type='iPhone12,5', hardware_model='d431ap', board_id=0x02, chip_id=0x8030,
                display_name='iPhone 11 Pro Max'),
    IRecvDevice(product_type='iPhone12,8', hardware_model='d79ap', board_id=0x10, chip_id=0x8030,
                display_name='iPhone SE (2020)'),
    IRecvDevice(product_type='iPhone13,1', hardware_model='d52gap', board_id=0x0A, chip_id=0x8101,
                display_name='iPhone 12 mini'),
    IRecvDevice(product_type='iPhone13,2', hardware_model='d53gap', board_id=0x0C, chip_id=0x8101,
                display_name='iPhone 12'),
    IRecvDevice(product_type='iPhone13,3', hardware_model='d53pap', board_id=0x0E, chip_id=0x8101,
                display_name='iPhone 12 Pro'),
    IRecvDevice(product_type='iPhone13,4', hardware_model='d54pap', board_id=0x08, chip_id=0x8101,
                display_name='iPhone 12 Pro Max'),
    # iPod
    IRecvDevice(product_type='iPod1,1', hardware_model='n45ap', board_id=0x02, chip_id=0x8900,
                display_name='iPod Touch (1st gen)'),
    IRecvDevice(product_type='iPod2,1', hardware_model='n72ap', board_id=0x00, chip_id=0x8720,
                display_name='iPod Touch (2nd gen)'),
    IRecvDevice(product_type='iPod3,1', hardware_model='n18ap', board_id=0x02, chip_id=0x8922,
                display_name='iPod Touch (3rd gen)'),
    IRecvDevice(product_type='iPod4,1', hardware_model='n81ap', board_id=0x08, chip_id=0x8930,
                display_name='iPod Touch (4th gen)'),
    IRecvDevice(product_type='iPod5,1', hardware_model='n78ap', board_id=0x00, chip_id=0x8942,
                display_name='iPod Touch (5th gen)'),
    IRecvDevice(product_type='iPod7,1', hardware_model='n102ap', board_id=0x10, chip_id=0x7000,
                display_name='iPod Touch (6th gen)'),
    IRecvDevice(product_type='iPod9,1', hardware_model='n112ap', board_id=0x16, chip_id=0x8010,
                display_name='iPod Touch (7th gen)'),
    # iPad
    IRecvDevice(product_type='iPad1,1', hardware_model='k48ap', board_id=0x02, chip_id=0x8930, display_name='iPad'),
    IRecvDevice(product_type='iPad2,1', hardware_model='k93ap', board_id=0x04, chip_id=0x8940,
                display_name='iPad 2 (WiFi)'),
    IRecvDevice(product_type='iPad2,2', hardware_model='k94ap', board_id=0x06, chip_id=0x8940,
                display_name='iPad 2 (GSM)'),
    IRecvDevice(product_type='iPad2,3', hardware_model='k95ap', board_id=0x02, chip_id=0x8940,
                display_name='iPad 2 (CDMA)'),
    IRecvDevice(product_type='iPad2,4', hardware_model='k93aap', board_id=0x06, chip_id=0x8942,
                display_name='iPad 2 (WiFi) R2 2012'),
    IRecvDevice(product_type='iPad2,5', hardware_model='p105ap', board_id=0x0a, chip_id=0x8942,
                display_name='iPad mini (WiFi)'),
    IRecvDevice(product_type='iPad2,6', hardware_model='p106ap', board_id=0x0c, chip_id=0x8942,
                display_name='iPad mini (GSM)'),
    IRecvDevice(product_type='iPad2,7', hardware_model='p107ap', board_id=0x0e, chip_id=0x8942,
                display_name='iPad mini (Global)'),
    IRecvDevice(product_type='iPad3,1', hardware_model='j1ap', board_id=0x00, chip_id=0x8945,
                display_name='iPad (3rd gen, WiFi)'),
    IRecvDevice(product_type='iPad3,2', hardware_model='j2ap', board_id=0x02, chip_id=0x8945,
                display_name='iPad (3rd gen, CDMA)'),
    IRecvDevice(product_type='iPad3,3', hardware_model='j2aap', board_id=0x04, chip_id=0x8945,
                display_name='iPad (3rd gen, GSM)'),
    IRecvDevice(product_type='iPad3,4', hardware_model='p101ap', board_id=0x00, chip_id=0x8955,
                display_name='iPad (4th gen, WiFi)'),
    IRecvDevice(product_type='iPad3,5', hardware_model='p102ap', board_id=0x02, chip_id=0x8955,
                display_name='iPad (4th gen, GSM)'),
    IRecvDevice(product_type='iPad3,6', hardware_model='p103ap', board_id=0x04, chip_id=0x8955,
                display_name='iPad (4th gen, Global)'),
    IRecvDevice(product_type='iPad4,1', hardware_model='j71ap', board_id=0x10, chip_id=0x8960,
                display_name='iPad Air (WiFi)'),
    IRecvDevice(product_type='iPad4,2', hardware_model='j72ap', board_id=0x12, chip_id=0x8960,
                display_name='iPad Air (Cellular)'),
    IRecvDevice(product_type='iPad4,3', hardware_model='j73ap', board_id=0x14, chip_id=0x8960,
                display_name='iPad Air (China)'),
    IRecvDevice(product_type='iPad4,4', hardware_model='j85ap', board_id=0x0a, chip_id=0x8960,
                display_name='iPad mini 2 (WiFi)'),
    IRecvDevice(product_type='iPad4,5', hardware_model='j86ap', board_id=0x0c, chip_id=0x8960,
                display_name='iPad mini 2 (Cellular)'),
    IRecvDevice(product_type='iPad4,6', hardware_model='j87ap', board_id=0x0e, chip_id=0x8960,
                display_name='iPad mini 2 (China)'),
    IRecvDevice(product_type='iPad4,7', hardware_model='j85map', board_id=0x32, chip_id=0x8960,
                display_name='iPad mini 3 (WiFi)'),
    IRecvDevice(product_type='iPad4,8', hardware_model='j86map', board_id=0x34, chip_id=0x8960,
                display_name='iPad mini 3 (Cellular)'),
    IRecvDevice(product_type='iPad4,9', hardware_model='j87map', board_id=0x36, chip_id=0x8960,
                display_name='iPad mini 3 (China)'),
    IRecvDevice(product_type='iPad5,1', hardware_model='j96ap', board_id=0x08, chip_id=0x7000,
                display_name='iPad mini 4 (WiFi)'),
    IRecvDevice(product_type='iPad5,2', hardware_model='j97ap', board_id=0x0A, chip_id=0x7000,
                display_name='iPad mini 4 (Cellular)'),
    IRecvDevice(product_type='iPad5,3', hardware_model='j81ap', board_id=0x06, chip_id=0x7001,
                display_name='iPad Air 2 (WiFi)'),
    IRecvDevice(product_type='iPad5,4', hardware_model='j82ap', board_id=0x02, chip_id=0x7001,
                display_name='iPad Air 2 (Cellular)'),
    IRecvDevice(product_type='iPad6,3', hardware_model='j127ap', board_id=0x08, chip_id=0x8001,
                display_name='iPad Pro 9.7-inch (WiFi)'),
    IRecvDevice(product_type='iPad6,4', hardware_model='j128ap', board_id=0x0a, chip_id=0x8001,
                display_name='iPad Pro 9.7-inch (Cellular)'),
    IRecvDevice(product_type='iPad6,7', hardware_model='j98aap', board_id=0x10, chip_id=0x8001,
                display_name='iPad Pro 12.9-inch (1st gen, WiFi)'),
    IRecvDevice(product_type='iPad6,8', hardware_model='j99aap', board_id=0x12, chip_id=0x8001,
                display_name='iPad Pro 12.9-inch (1st gen, Cellular)'),
    IRecvDevice(product_type='iPad6,11', hardware_model='j71sap', board_id=0x10, chip_id=0x8000,
                display_name='iPad (5th gen, WiFi)'),
    IRecvDevice(product_type='iPad6,11', hardware_model='j71tap', board_id=0x10, chip_id=0x8003,
                display_name='iPad (5th gen, WiFi)'),
    IRecvDevice(product_type='iPad6,12', hardware_model='j72sap', board_id=0x12, chip_id=0x8000,
                display_name='iPad (5th gen, Cellular)'),
    IRecvDevice(product_type='iPad6,12', hardware_model='j72tap', board_id=0x12, chip_id=0x8003,
                display_name='iPad (5th gen, Cellular)'),
    IRecvDevice(product_type='iPad7,1', hardware_model='j120ap', board_id=0x0C, chip_id=0x8011,
                display_name='iPad Pro 12.9-inch (2nd gen, WiFi)'),
    IRecvDevice(product_type='iPad7,2', hardware_model='j121ap', board_id=0x0E, chip_id=0x8011,
                display_name='iPad Pro 12.9-inch (2nd gen, Cellular)'),
    IRecvDevice(product_type='iPad7,3', hardware_model='j207ap', board_id=0x04, chip_id=0x8011,
                display_name='iPad Pro 10.5-inch (WiFi)'),
    IRecvDevice(product_type='iPad7,4', hardware_model='j208ap', board_id=0x06, chip_id=0x8011,
                display_name='iPad Pro 10.5-inch (Cellular)'),
    IRecvDevice(product_type='iPad7,5', hardware_model='j71bap', board_id=0x18, chip_id=0x8010,
                display_name='iPad (6th gen, WiFi)'),
    IRecvDevice(product_type='iPad7,6', hardware_model='j72bap', board_id=0x1A, chip_id=0x8010,
                display_name='iPad (6th gen, Cellular)'),
    IRecvDevice(product_type='iPad7,11', hardware_model='j171ap', board_id=0x1C, chip_id=0x8010,
                display_name='iPad (7th gen, WiFi)'),
    IRecvDevice(product_type='iPad7,12', hardware_model='j172ap', board_id=0x1E, chip_id=0x8010,
                display_name='iPad (7th gen, Cellular)'),
    IRecvDevice(product_type='iPad8,1', hardware_model='j317ap', board_id=0x0C, chip_id=0x8027,
                display_name='iPad Pro 11-inch (1st gen, WiFi)'),
    IRecvDevice(product_type='iPad8,2', hardware_model='j317xap', board_id=0x1C, chip_id=0x8027,
                display_name='iPad Pro 11-inch (1st gen, WiFi, 1TB)'),
    IRecvDevice(product_type='iPad8,3', hardware_model='j318ap', board_id=0x0E, chip_id=0x8027,
                display_name='iPad Pro 11-inch (1st gen, Cellular)'),
    IRecvDevice(product_type='iPad8,4', hardware_model='j318xap', board_id=0x1E, chip_id=0x8027,
                display_name='iPad Pro 11-inch (1st gen, Cellular, 1TB)'),
    IRecvDevice(product_type='iPad8,5', hardware_model='j320ap', board_id=0x08, chip_id=0x8027,
                display_name='iPad Pro 12.9-inch (3rd gen, WiFi)'),
    IRecvDevice(product_type='iPad8,6', hardware_model='j320xap', board_id=0x18, chip_id=0x8027,
                display_name='iPad Pro 12.9-inch (3rd gen, WiFi, 1TB)'),
    IRecvDevice(product_type='iPad8,7', hardware_model='j321ap', board_id=0x0A, chip_id=0x8027,
                display_name='iPad Pro 12.9-inch (3rd gen, Cellular)'),
    IRecvDevice(product_type='iPad8,8', hardware_model='j321xap', board_id=0x1A, chip_id=0x8027,
                display_name='iPad Pro 12.9-inch (3rd gen, Cellular, 1TB)'),
    IRecvDevice(product_type='iPad8,9', hardware_model='j417ap', board_id=0x3C, chip_id=0x8027,
                display_name='iPad Pro 11-inch (2nd gen, WiFi)'),
    IRecvDevice(product_type='iPad8,10', hardware_model='j418ap', board_id=0x3E, chip_id=0x8027,
                display_name='iPad Pro 11-inch (2nd gen, Cellular)'),
    IRecvDevice(product_type='iPad8,11', hardware_model='j420ap', board_id=0x38, chip_id=0x8027,
                display_name='iPad Pro 12.9-inch (4th gen, WiFi)'),
    IRecvDevice(product_type='iPad8,12', hardware_model='j421ap', board_id=0x3A, chip_id=0x8027,
                display_name='iPad Pro 12.9-inch (4th gen, Cellular)'),
    IRecvDevice(product_type='iPad11,1', hardware_model='j210ap', board_id=0x14, chip_id=0x8020,
                display_name='iPad mini (5th gen, WiFi)'),
    IRecvDevice(product_type='iPad11,2', hardware_model='j211ap', board_id=0x16, chip_id=0x8020,
                display_name='iPad mini (5th gen, Cellular)'),
    IRecvDevice(product_type='iPad11,3', hardware_model='j217ap', board_id=0x1C, chip_id=0x8020,
                display_name='iPad Air (3rd gen, WiFi)'),
    IRecvDevice(product_type='iPad11,4', hardware_model='j218ap', board_id=0x1E, chip_id=0x8020,
                display_name='iPad Air (3rd gen, Celluar)'),
    IRecvDevice(product_type='iPad11,6', hardware_model='j171aap', board_id=0x24, chip_id=0x8020,
                display_name='iPad (8th gen, WiFi)'),
    IRecvDevice(product_type='iPad11,7', hardware_model='j172aap', board_id=0x26, chip_id=0x8020,
                display_name='iPad (8th gen, Celluar)'),
    IRecvDevice(product_type='iPad13,1', hardware_model='j307ap', board_id=0x04, chip_id=0x8101,
                display_name='iPad Air (4th gen, WiFi)'),
    IRecvDevice(product_type='iPad13,2', hardware_model='j308ap', board_id=0x06, chip_id=0x8101,
                display_name='iPad Air (4th gen, Celluar)'),
    IRecvDevice(product_type='iPad13,4', hardware_model='j517ap', board_id=0x08, chip_id=0x8103,
                display_name='iPad Pro 11-inch (3rd gen, WiFi)'),
    IRecvDevice(product_type='iPad13,5', hardware_model='j517xap', board_id=0x0A, chip_id=0x8103,
                display_name='iPad Pro 11-inch (3rd gen, WiFi, 2TB)'),
    IRecvDevice(product_type='iPad13,6', hardware_model='j518ap', board_id=0x0C, chip_id=0x8103,
                display_name='iPad Pro 11-inch (3rd gen, Cellular)'),
    IRecvDevice(product_type='iPad13,7', hardware_model='j518xap', board_id=0x0E, chip_id=0x8103,
                display_name='iPad Pro 11-inch (3rd gen, Celluar, 2TB)'),
    IRecvDevice(product_type='iPad13,8', hardware_model='j522ap', board_id=0x18, chip_id=0x8103,
                display_name='iPad Pro 12.9-inch (5th gen, WiFi)'),
    IRecvDevice(product_type='iPad13,9', hardware_model='j522xap', board_id=0x1A, chip_id=0x8103,
                display_name='iPad Pro 12.9-inch (5th gen, WiFi, 2TB)'),
    IRecvDevice(product_type='iPad13,10', hardware_model='j523ap', board_id=0x1C, chip_id=0x8103,
                display_name='iPad Pro 12.9-inch (5th gen, Celluar)'),
    IRecvDevice(product_type='iPad13,11', hardware_model='j523xap', board_id=0x1E, chip_id=0x8103,
                display_name='iPad Pro 12.9-inch (5th gen, Celluar, 2TB)'),
    # Apple TV
    IRecvDevice(product_type='AppleTV2,1', hardware_model='k66ap', board_id=0x10, chip_id=0x8930,
                display_name='Apple TV 2'),
    IRecvDevice(product_type='AppleTV3,1', hardware_model='j33ap', board_id=0x08, chip_id=0x8942,
                display_name='Apple TV 3'),
    IRecvDevice(product_type='AppleTV3,2', hardware_model='j33iap', board_id=0x00, chip_id=0x8947,
                display_name='Apple TV 3 (2013)'),
    IRecvDevice(product_type='AppleTV5,3', hardware_model='j42dap', board_id=0x34, chip_id=0x7000,
                display_name='Apple TV 4'),
    IRecvDevice(product_type='AppleTV6,2', hardware_model='j105aap', board_id=0x02, chip_id=0x8011,
                display_name='Apple TV 4K'),
    IRecvDevice(product_type='AppleTV11,1', hardware_model='j305ap', board_id=0x08, chip_id=0x8020,
                display_name='Apple TV 4K (2nd gen)'),
    # HomePod
    IRecvDevice(product_type='AudioAccessory1,1', hardware_model='b238aap', board_id=0x38, chip_id=0x7000,
                display_name='HomePod'),
    IRecvDevice(product_type='AudioAccessory1,2', hardware_model='b238ap', board_id=0x1A, chip_id=0x7000,
                display_name='HomePod'),
    IRecvDevice(product_type='AudioAccessory5,1', hardware_model='b520ap', board_id=0x22, chip_id=0x8006,
                display_name='HomePod mini'),
    # Apple Watch
    IRecvDevice(product_type='Watch1,1', hardware_model='n27aap', board_id=0x02, chip_id=0x7002,
                display_name='Apple Watch 38mm (1st gen)'),
    IRecvDevice(product_type='Watch1,2', hardware_model='n28aap', board_id=0x04, chip_id=0x7002,
                display_name='Apple Watch 42mm (1st gen)'),
    IRecvDevice(product_type='Watch2,6', hardware_model='n27dap', board_id=0x02, chip_id=0x8002,
                display_name='Apple Watch Series 1 (38mm)'),
    IRecvDevice(product_type='Watch2,7', hardware_model='n28dap', board_id=0x04, chip_id=0x8002,
                display_name='Apple Watch Series 1 (42mm)'),
    IRecvDevice(product_type='Watch2,3', hardware_model='n74ap', board_id=0x0C, chip_id=0x8002,
                display_name='Apple Watch Series 2 (38mm)'),
    IRecvDevice(product_type='Watch2,4', hardware_model='n75ap', board_id=0x0E, chip_id=0x8002,
                display_name='Apple Watch Series 2 (42mm)'),
    IRecvDevice(product_type='Watch3,1', hardware_model='n111sap', board_id=0x1C, chip_id=0x8004,
                display_name='Apple Watch Series 3 (38mm Cellular)'),
    IRecvDevice(product_type='Watch3,2', hardware_model='n111bap', board_id=0x1E, chip_id=0x8004,
                display_name='Apple Watch Series 3 (42mm Cellular)'),
    IRecvDevice(product_type='Watch3,3', hardware_model='n121sap', board_id=0x18, chip_id=0x8004,
                display_name='Apple Watch Series 3 (38mm)'),
    IRecvDevice(product_type='Watch3,4', hardware_model='n121bap', board_id=0x1A, chip_id=0x8004,
                display_name='Apple Watch Series 3 (42mm)'),
    IRecvDevice(product_type='Watch4,1', hardware_model='n131sap', board_id=0x08, chip_id=0x8006,
                display_name='Apple Watch Series 4 (40mm)'),
    IRecvDevice(product_type='Watch4,2', hardware_model='n131bap', board_id=0x0A, chip_id=0x8006,
                display_name='Apple Watch Series 4 (44mm)'),
    IRecvDevice(product_type='Watch4,3', hardware_model='n141sap', board_id=0x0C, chip_id=0x8006,
                display_name='Apple Watch Series 4 (40mm Cellular)'),
    IRecvDevice(product_type='Watch4,4', hardware_model='n141bap', board_id=0x0E, chip_id=0x8006,
                display_name='Apple Watch Series 4 (44mm Cellular)'),
    IRecvDevice(product_type='Watch5,1', hardware_model='n144sap', board_id=0x10, chip_id=0x8006,
                display_name='Apple Watch Series 5 (40mm)'),
    IRecvDevice(product_type='Watch5,2', hardware_model='n144bap', board_id=0x12, chip_id=0x8006,
                display_name='Apple Watch Series 5 (44mm)'),
    IRecvDevice(product_type='Watch5,3', hardware_model='n146sap', board_id=0x14, chip_id=0x8006,
                display_name='Apple Watch Series 5 (40mm Cellular)'),
    IRecvDevice(product_type='Watch5,4', hardware_model='n146bap', board_id=0x16, chip_id=0x8006,
                display_name='Apple Watch Series 5 (44mm Cellular)'),
    IRecvDevice(product_type='Watch5,9', hardware_model='n140sap', board_id=0x28, chip_id=0x8006,
                display_name='Apple Watch SE (40mm)'),
    IRecvDevice(product_type='Watch5,10', hardware_model='n140bap', board_id=0x2A, chip_id=0x8006,
                display_name='Apple Watch SE (44mm)'),
    IRecvDevice(product_type='Watch5,11', hardware_model='n142sap', board_id=0x2C, chip_id=0x8006,
                display_name='Apple Watch SE (40mm Cellular)'),
    IRecvDevice(product_type='Watch5,12', hardware_model='n142bap', board_id=0x2E, chip_id=0x8006,
                display_name='Apple Watch SE (44mm Cellular)'),
    IRecvDevice(product_type='Watch6,1', hardware_model='n157sap', board_id=0x08, chip_id=0x8301,
                display_name='Apple Watch Series 6 (40mm)'),
    IRecvDevice(product_type='Watch6,2', hardware_model='n157bap', board_id=0x0A, chip_id=0x8301,
                display_name='Apple Watch Series 6 (44mm)'),
    IRecvDevice(product_type='Watch6,3', hardware_model='n158sap', board_id=0x0C, chip_id=0x8301,
                display_name='Apple Watch Series 6 (40mm Cellular)'),
    IRecvDevice(product_type='Watch6,4', hardware_model='n158bap', board_id=0x0E, chip_id=0x8301,
                display_name='Apple Watch Series 6 (44mm Cellular)'),
    # Apple Silicon Macs
    IRecvDevice(product_type='ADP3,2', hardware_model='j273aap', board_id=0x42, chip_id=0x8027,
                display_name='Developer Transition Kit'),
    IRecvDevice(product_type='Macmini9,1', hardware_model='j274ap', board_id=0x22, chip_id=0x8103,
                display_name='Mac mini (M1, 2020)'),
    IRecvDevice(product_type='MacBookPro17,1', hardware_model='j293ap', board_id=0x24, chip_id=0x8103,
                display_name='MacBook Pro (M1, 2020)'),
    IRecvDevice(product_type='MacBookAir10,1', hardware_model='j313ap', board_id=0x26, chip_id=0x8103,
                display_name='MacBook Air (M1, 2020)'),
    IRecvDevice(product_type='iMac21,1', hardware_model='j456ap', board_id=0x28, chip_id=0x8103,
                display_name='iMac 24-inch (M1, Two Ports, 2021)'),
    IRecvDevice(product_type='iMac21,2', hardware_model='j457ap', board_id=0x2A, chip_id=0x8103,
                display_name='iMac 24-inch (M1, Four Ports, 2021)'),
    # Apple T2 Coprocessor
    IRecvDevice(product_type='iBridge2,1', hardware_model='j137ap', board_id=0x0A, chip_id=0x8012,
                display_name='Apple T2 iMacPro1,1 (j137)'),
    IRecvDevice(product_type='iBridge2,3', hardware_model='j680ap', board_id=0x0B, chip_id=0x8012,
                display_name='Apple T2 MacBookPro15,1 (j680)'),
    IRecvDevice(product_type='iBridge2,4', hardware_model='j132ap', board_id=0x0C, chip_id=0x8012,
                display_name='Apple T2 MacBookPro15,2 (j132)'),
    IRecvDevice(product_type='iBridge2,5', hardware_model='j174ap', board_id=0x0E, chip_id=0x8012,
                display_name='Apple T2 Macmini8,1 (j174)'),
    IRecvDevice(product_type='iBridge2,6', hardware_model='j160ap', board_id=0x0F, chip_id=0x8012,
                display_name='Apple T2 MacPro7,1 (j160)'),
    IRecvDevice(product_type='iBridge2,7', hardware_model='j780ap', board_id=0x07, chip_id=0x8012,
                display_name='Apple T2 MacBookPro15,3 (j780)'),
    IRecvDevice(product_type='iBridge2,8', hardware_model='j140kap', board_id=0x17, chip_id=0x8012,
                display_name='Apple T2 MacBookAir8,1 (j140k)'),
    IRecvDevice(product_type='iBridge2,10', hardware_model='j213ap', board_id=0x18, chip_id=0x8012,
                display_name='Apple T2 MacBookPro15,4 (j213)'),
    IRecvDevice(product_type='iBridge2,12', hardware_model='j140aap', board_id=0x37, chip_id=0x8012,
                display_name='Apple T2 MacBookAir8,2 (j140a)'),
    IRecvDevice(product_type='iBridge2,14', hardware_model='j152fap', board_id=0x3A, chip_id=0x8012,
                display_name='Apple T2 MacBookPro16,1 (j152f)'),
    IRecvDevice(product_type='iBridge2,15', hardware_model='j230kap', board_id=0x3F, chip_id=0x8012,
                display_name='Apple T2 MacBookAir9,1 (j230k)'),
    IRecvDevice(product_type='iBridge2,16', hardware_model='j214kap', board_id=0x3E, chip_id=0x8012,
                display_name='Apple T2 MacBookPro16,2 (j214k)'),
    IRecvDevice(product_type='iBridge2,19', hardware_model='j185ap', board_id=0x22, chip_id=0x8012,
                display_name='Apple T2 iMac20,1 (j185)'),
    IRecvDevice(product_type='iBridge2,20', hardware_model='j185fap', board_id=0x23, chip_id=0x8012,
                display_name='Apple T2 iMac20,2 (j185f)'),
    IRecvDevice(product_type='iBridge2,21', hardware_model='j223ap', board_id=0x3B, chip_id=0x8012,
                display_name='Apple T2 MacBookPro16,3 (j223)'),
    IRecvDevice(product_type='iBridge2,22', hardware_model='j215ap', board_id=0x38, chip_id=0x8012,
                display_name='Apple T2 MacBookPro16,4 (j215)'),
)



#!/usr/bin/env python3
import logging
import plistlib
import struct
import tempfile
from datetime import datetime
from io import BytesIO
from tarfile import TarFile

from construct import Struct, Bytes, Int32ul, Optional, Enum, Byte, Adapter, Int16ul, this, Computed, \
    RepeatUntil
from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient

CHUNK_SIZE = 4096
TIME_FORMAT = '%H:%M:%S'
SYSLOG_LINE_SPLITTER = '\n\x00'


class TimestampAdapter(Adapter):
    def _decode(self, obj, context, path):
        return datetime.fromtimestamp(obj.seconds + (obj.microseconds / 1000000))

    def _encode(self, obj, context, path):
        return list(map(int, obj.split(".")))


timestamp_t = Struct(
    'seconds' / Int32ul,
    Bytes(4),
    'microseconds' / Int32ul
)


def try_decode(s):
    try:
        return s.decode('utf8')
    except UnicodeDecodeError:
        return s


syslog_t = Struct(
    Bytes(9),
    'pid' / Int32ul,
    Bytes(42),
    'timestamp' / TimestampAdapter(timestamp_t),
    Bytes(1),
    'level' / Enum(Byte, Notice=0, Info=0x01, Debug=0x02, Error=0x10, Fault=0x11),
    Bytes(38),
    'image_name_size' / Int16ul,
    'message_size' / Int16ul,
    Bytes(6),
    '_subsystem_size' / Int32ul,
    '_category_size' / Int32ul,
    Bytes(4),
    '_filename' / RepeatUntil(lambda x, lst, ctx: lst[-1] == 0, Byte),
    'filename' / Computed(lambda ctx: try_decode(bytearray(ctx._filename[:-1]))),
    '_image_name' / Bytes(this.image_name_size),
    'image_name' / Computed(lambda ctx: try_decode(ctx._image_name[:-1])),
    '_message' / Bytes(this.message_size),
    'message' / Computed(lambda ctx: try_decode(ctx._message[:-1])),
    'label' / Optional(Struct(
        '_subsystem' / Bytes(this._._subsystem_size),
        'subsystem' / Computed(lambda ctx: try_decode(ctx._subsystem[:-1])),
        '_category' / Bytes(this._._category_size),
        'category' / Computed(lambda ctx: try_decode(ctx._category[:-1])),
    )),
)


class OsTraceService(object):
    SERVICE_NAME = 'com.apple.os_trace_relay'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.c = self.lockdown.start_service(self.SERVICE_NAME)

    def get_pid_list(self):
        self.c.send_plist({'Request': 'PidList'})

        # ignore first received unknown byte
        self.c.recvall(1)

        response = self.c.recv_prefixed()
        return plistlib.loads(response)

    def create_archive(self, out: BytesIO):
        self.c.send_plist({'Request': 'CreateArchive'})

        assert 1 == self.c.recvall(1)[0]

        assert plistlib.loads(self.c.recv_prefixed()).get('Status') == 'RequestSuccessful', 'Invalid status'

        while True:
            try:
                assert 3 == self.c.recvall(1)[0], 'invalid magic'
            except ConnectionAbortedError:
                break
            out.write(self.c.recv_prefixed(endianity='<'))

    def collect(self, out: str):
        """
        Collect the system logs into a .logarchive that can be viewed later with tools such as log or Console.
        """
        with tempfile.NamedTemporaryFile() as tar:
            self.create_archive(tar)
            TarFile(tar.name).extractall(out)

    def syslog(self, pid=-1):
        self.c.send_plist({'Request': 'StartActivity', 'MessageFilter': 65535, 'Pid': pid, 'StreamFlags': 60})

        length_length, = struct.unpack('<I', self.c.recvall(4))
        length = int(self.c.recvall(length_length)[::-1].hex(), 16)
        response = plistlib.loads(self.c.recvall(length))

        if response.get('Status') != 'RequestSuccessful':
            raise PyMobileDevice3Exception(f'got invalid response: {response}')

        while True:
            assert b'\x02' == self.c.recvall(1)
            length, = struct.unpack('<I', self.c.recvall(4))
            line = self.c.recvall(length)
            entry = syslog_t.parse(line)
            yield entry



import logging
import time
from io import BytesIO

from usb import USBError

from pymobiledevice3.irecv import IRecv, Mode
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.restore.device import Device
from pymobiledevice3.restore.ipsw.ipsw import IPSW
from pymobiledevice3.restore.tss import TSSRequest, TSSResponse


class Recovery:
    def __init__(self, ipsw: BytesIO, device: Device, tss: dict = None,
                 offline=False, behavior='Update'):
        self.ipsw = IPSW(ipsw)
        self.device = device
        self.offline = offline
        self.tss = TSSResponse(tss) if tss is not None else None

        if not self.device.is_image4_supported:
            raise NotImplementedError('is_image4_supported is False')

        logging.info(f'connected device: <ecid: {self.device.ecid} hardware_model: {self.device.hardware_model} '
                     f'image4-support: {self.device.is_image4_supported}>')

        logging.debug('scanning BuildManifest.plist for the correct BuildIdentity')
        self.build_identity = self.ipsw.build_manifest.get_build_identity(self.device.hardware_model, behavior)
        self.restore_boot_args = None

    def reconnect_irecv(self, is_recovery=None):
        logging.debug('waiting for device to reconnect...')
        self.device.irecv = IRecv(ecid=self.device.ecid, is_recovery=is_recovery)
        logging.debug(f'connected mode: {self.device.irecv.mode}')

    def get_preboard_manifest(self):
        overrides = {
            '@APTicket': True,
            'ApProductionMode': 0,
            'ApSecurityDomain': 0,
        }

        parameters = {
            'ApProductionMode': False,
            'ApSecurityMode': False,
            'ApSupportsImg4': True,
        }

        self.build_identity.populate_tss_request_parameters(parameters)

        tss = TSSRequest(offline=self.offline)
        tss.add_common_tags(parameters, overrides)

        parameters['_OnlyFWComponents'] = True

        tss.add_ap_tags(parameters)

        return tss.img4_create_local_manifest(build_identity=self.build_identity)

    def get_tss_response(self):
        # populate parameters
        parameters = dict()

        parameters['ApECID'] = self.device.ecid
        if self.device.ap_nonce is not None:
            parameters['ApNonce'] = self.device.ap_nonce

        if self.device.sep_nonce is not None:
            parameters['ApSepNonce'] = self.device.sep_nonce

        parameters['ApProductionMode'] = True

        if self.device.is_image4_supported:
            parameters['ApSecurityMode'] = True
            parameters['ApSupportsImg4'] = True
        else:
            parameters['ApSupportsImg4'] = False

        self.build_identity.populate_tss_request_parameters(parameters)

        tss = TSSRequest(offline=self.offline)
        tss.add_common_tags(parameters)
        tss.add_ap_tags(parameters)

        # add personalized parameters
        if self.device.is_image4_supported:
            tss.add_ap_img4_tags(parameters)
        else:
            tss.add_ap_img3_tags(parameters)

        # normal mode; request baseband ticket as well
        if self.device.lockdown is not None:
            pinfo = self.device.lockdown.preflight_info
            if pinfo:
                logging.debug('adding preflight info')

                node = pinfo.get('Nonce')
                if node is not None:
                    parameters['BbNonce'] = node

                node = pinfo.get('ChipID')
                if node is not None:
                    parameters['BbChipID'] = node

                node = pinfo.get('CertID')
                if node is not None:
                    parameters['BbGoldCertId'] = node

                node = pinfo.get('ChipSerialNo')
                if node is not None:
                    parameters['BbSNUM'] = node

                tss.add_baseband_tags(parameters)

                euiccchipid = pinfo.get('EUICCChipID')
                if euiccchipid:
                    logging.debug('adding EUICCChipID info')
                    parameters['eUICC,ChipID'] = euiccchipid

                    if euiccchipid >= 5:
                        node = pinfo.get('EUICCCSN')
                        if node is not None:
                            parameters['eUICC,EID'] = node

                        node = pinfo.get('EUICCCertIdentifier')
                        if node is not None:
                            parameters['eUICC,RootKeyIdentifier'] = node

                        node = pinfo.get('EUICCGoldNonce')
                        if node is not None:
                            parameters['EUICCGoldNonce'] = node

                        node = pinfo.get('EUICCMainNonce')
                        if node is not None:
                            parameters['EUICCMainNonce'] = node

                        tss.add_vinyl_tags(parameters)

        # send request and grab response
        return tss.send_receive()

    def fetch_tss_record(self):
        if self.ipsw.build_manifest.build_major > 8:
            if self.device.ap_nonce is None:
                # the first nonce request with older firmware releases can fail and it's OK
                logging.info('NOTE: Unable to get nonce from device')

        self.tss = self.get_tss_response()

        if self.ipsw.build_manifest.build_major >= 20:
            raise NotImplementedError('not yet supported')

        return self.tss

    def send_component(self, name):
        self.device.irecv.send_buffer(self.build_identity.get_component(name, tss=self.tss).personalized_data)

    def send_component_and_command(self, name, command):
        self.send_component(name)
        self.device.irecv.send_command(command)

    def send_ibec(self):
        component = 'iBEC'
        self.send_component(component)
        self.device.irecv.send_command('go')
        self.device.irecv.ctrl_transfer(0x21, 1)

    def send_applelogo(self):
        component = 'RestoreLogo'

        if not self.build_identity.has_component(component):
            return

        self.send_component(component)
        self.device.irecv.send_command('setpicture 4')
        self.device.irecv.send_command('bgcolor 0 0 0')

    def send_loaded_by_iboot(self):
        manifest = self.build_identity['Manifest']
        for key, node in manifest.items():
            iboot = node['Info'].get('IsLoadedByiBoot', False)
            iboot_stg1 = node['Info'].get('IsLoadedByiBootStage1', False)

            assert isinstance(iboot, bool)
            assert isinstance(iboot_stg1, bool)

            if iboot and not iboot_stg1:
                logging.debug(f'{key} is loaded by iBoot')
                self.send_component_and_command(key, 'firmware')

    def send_ramdisk(self):
        component = 'RestoreRamDisk'
        ramdisk_size = self.device.irecv.getenv('ramdisk-size')
        logging.info(f'ramdisk-size: {ramdisk_size}')

        self.send_component(component)
        ramdisk_delay = self.device.irecv.getenv('ramdisk-delay')
        logging.info(f'ramdisk-delay: {ramdisk_delay}')

        self.device.irecv.send_command('ramdisk')

        time.sleep(2)

    def send_kernelcache(self):
        component = 'RestoreKernelCache'

        self.send_component(component)
        try:
            self.device.irecv.ctrl_transfer(0x21, 1)
        except USBError:
            pass

        if self.restore_boot_args:
            self.device.irecv.send_command(f'setenv boot-args {self.restore_boot_args}')

        try:
            self.device.irecv.send_command('bootx')
        except USBError:
            pass

    def set_autoboot(self, enable: bool):
        self.device.irecv.set_autoboot(enable)

    def enter_restore(self):
        if self.ipsw.build_manifest.build_major >= 8:
            self.restore_boot_args = 'rd=md0 nand-enable-reformat=1 -progress'
        elif self.ipsw.build_manifest.build_major >= 20:
            self.restore_boot_args = 'rd=md0 nand-enable-reformat=1 -progress -restore'

        # upload data to make device boot restore mode

        # Recovery Mode Environment:
        build_version = self.device.irecv.getenv('build-version')
        logging.info(f'iBoot build-version={build_version}')

        build_style = self.device.irecv.getenv('build-style')
        logging.info(f'iBoot build-style={build_style}')

        radio_error = self.device.irecv.getenv('radio-error')
        if radio_error:
            radio_error = int(radio_error)
            logging.info(f'radio-error: {radio_error}')
            radio_error_string = self.device.irecv.getenv('radio-error-string')
            if radio_error_string:
                logging.info(f'radio-error-string: {radio_error_string}')

        self.set_autoboot(False)

        # send logo and show it
        self.send_applelogo()

        # send components loaded by iBoot
        self.send_loaded_by_iboot()

        # send ramdisk and run it
        self.send_ramdisk()

        # send devicetree and load it
        self.send_component_and_command('RestoreDeviceTree', 'devicetree')

        if 'RestoreSEP' in self.build_identity['Manifest']:
            # send rsepfirmware and load it
            self.send_component_and_command('RestoreSEP', 'rsepfirmware')

        self.send_kernelcache()

    def dfu_enter_recovery(self):
        self.send_component('iBSS')
        self.reconnect_irecv(is_recovery=True)

        if self.build_identity.build_manifest.build_major > 8:
            old_nonce = self.device.irecv.ap_nonce

            # reconnect
            self.reconnect_irecv()
            nonce = self.device.irecv.ap_nonce

            if old_nonce != nonce:
                # Welcome iOS5. We have to re-request the TSS with our nonce.
                self.tss = self.get_tss_response()

        self.device.irecv.set_configuration(1)

    def boot_ramdisk(self):
        if self.tss is None:
            logging.info('fetching TSS record')
            self.fetch_tss_record()

        if self.device.irecv and self.device.irecv.mode == Mode.DFU_MODE:
            # device is currently in DFU mode, place it into recovery mode
            self.dfu_enter_recovery()

            # Now, before sending iBEC, we must send necessary firmwares on new versions.
            if self.build_identity.build_manifest.build_major >= 20:
                # Without this empty policy file & its special signature, iBEC won't start.
                raise NotImplementedError()

        if self.device.lockdown:
            # normal mode
            logging.info('going into Recovery')

            # in case lockdown has disconnected while waiting for a ticket
            self.device.lockdown = LockdownClient(udid=self.device.lockdown.udid)
            self.device.lockdown.enter_recovery()

            self.device.lockdown = None
            self.device.irecv = IRecv(self.device.ecid)

        self.reconnect_irecv()
        ecid = self.device.irecv._device_info['ECID']
        logging.debug(f'ECID: {ecid}')

        logging.info('device booted into recovery')

        # now we load the iBEC
        try:
            self.send_ibec()
        except USBError:
            pass

        self.reconnect_irecv()

        # now finally do the magic to put the device into restore mode
        self.enter_restore()



from pathlib import Path

import pytest

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.exceptions import DvtDirListError, AlreadyMountedError
from pymobiledevice3.services.dvt.instruments.application_listing import ApplicationListing
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo
from pymobiledevice3.services.dvt.instruments.process_control import ProcessControl
from pymobiledevice3.services.mobile_image_mounter import MobileImageMounterService

DEVICE_SUPPORT = Path('/Applications/Xcode.app/Contents/Developer/Platforms/iPhoneOS.platform/DeviceSupport')
IMAGE_TYPE = 'Developer'


@pytest.fixture(scope='module', autouse=True)
def mount_developer_disk_image():
    mounter = MobileImageMounterService(lockdown=LockdownClient())

    image_path = DEVICE_SUPPORT / mounter.lockdown.sanitized_ios_version / 'DeveloperDiskImage.dmg'
    signature = image_path.with_suffix('.dmg.signature').read_bytes()
    mounter.upload_image('Developer', image_path.read_bytes(), signature)
    try:
        mounter.mount(IMAGE_TYPE, signature)
    except AlreadyMountedError:
        pass


def get_process_data(lockdown, name):
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        processes = DeviceInfo(dvt).proclist()
    return [process for process in processes if process['name'] == name][0]


def test_ls(lockdown):
    """
    Test listing a directory.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        ls = set(DeviceInfo(dvt).ls('/'))
    assert {'usr', 'bin', 'etc', 'var', 'private', 'lib', 'Applications', 'Developer'} <= ls


def test_ls_failure(lockdown):
    """
    Test listing a directory.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        with pytest.raises(DvtDirListError):
            DeviceInfo(dvt).ls('Directory that does not exist')


def test_proclist(lockdown):
    """
    Test listing processes.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    lockdownd = get_process_data(lockdown, 'lockdownd')
    assert lockdownd['realAppName'] == '/usr/libexec/lockdownd'
    assert not lockdownd['isApplication']


def test_applist(lockdown):
    """
    Test listing applications.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        apps = ApplicationListing(dvt).applist()

    safari = [app for app in apps if app['BundlePath'] == '/Applications/MobileSafari.app'][0]
    assert safari['DisplayName'] == 'Safari'
    assert safari['CFBundleIdentifier'] == 'com.apple.mobilesafari'
    assert safari['ExecutableName'] == 'MobileSafari'
    assert not safari['Placeholder']
    assert safari['Type'] == 'System'


def test_kill(lockdown):
    """
    Test killing a process.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    aggregated = get_process_data(lockdown, 'aggregated')

    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        ProcessControl(dvt).kill(aggregated['pid'])

    aggregated_after_kill = get_process_data(lockdown, 'aggregated')

    assert aggregated['startDate'] < aggregated_after_kill['startDate']


def test_launch(lockdown):
    """
    Test launching a process.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        pid = ProcessControl(dvt).launch('com.apple.mobilesafari')
        assert pid
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        for process in DeviceInfo(dvt).proclist():
            if pid == process['pid']:
                assert process['name'] == 'MobileSafari'


def test_system_information(lockdown):
    """
    Test getting system information.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        system_info = DeviceInfo(dvt).system_information()
    assert '_deviceDescription' in system_info and system_info['_deviceDescription'].startswith('Build Version')


def test_hardware_information(lockdown):
    """
    Test getting hardware information.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        hardware_info = DeviceInfo(dvt).hardware_information()
    assert hardware_info['numberOfCpus'] > 0


def test_network_information(lockdown):
    """
    Test getting network information.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        network_info = DeviceInfo(dvt).network_information()
    assert network_info['lo0'] == 'Loopback'



import logging
import os
import plistlib
import traceback

import IPython
import click
from pygments import highlight, lexers, formatters
from pymobiledevice3.cli.cli_common import print_json
from pymobiledevice3.exceptions import IncorrectModeError
from pymobiledevice3.irecv import IRecv
from pymobiledevice3.lockdown import list_devices, LockdownClient
from pymobiledevice3.restore.device import Device
from pymobiledevice3.restore.recovery import Recovery
from pymobiledevice3.restore.restore import Restore

SHELL_USAGE = """
# use `irecv` variable to access Restore mode API
# for example:
print(irecv.getenv('build-version'))
"""


class Command(click.Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params[:0] = [
            click.Option(('device', '--ecid'), type=click.INT, callback=self.device),
        ]

    @staticmethod
    def device(ctx, param, value):
        if '_PYMOBILEDEVICE3_COMPLETE' in os.environ:
            # prevent lockdown connection establishment when in autocomplete mode
            return

        ecid = value
        logging.debug('searching among connected devices via lockdownd')
        for udid in list_devices():
            try:
                lockdown = LockdownClient(udid=udid)
            except IncorrectModeError:
                continue
            if (ecid is None) or (lockdown.ecid == value):
                logging.debug('found device')
                return lockdown
            else:
                continue
        logging.debug(f'waiting for device to be available in Recovery mode')
        return IRecv(ecid=ecid)


@click.group()
def cli():
    """ cli """
    pass


@cli.group()
def restore():
    """ restore options """
    pass


@restore.command('shell', cls=Command)
def restore_shell(device):
    """ create an IPython shell for interacting with iBoot """
    IPython.embed(
        header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.TerminalTrueColorFormatter(style='native')),
        user_ns={
            'irecv': device,
        })


@restore.command('enter', cls=Command)
def restore_enter(lockdown):
    """ enter Recovery mode """
    lockdown.enter_recovery()


@restore.command('exit')
def restore_exit():
    """ exit Recovery mode """
    irecv = IRecv()
    irecv.set_autoboot(True)
    irecv.reboot()


@restore.command('tss', cls=Command)
@click.argument('ipsw', type=click.File('rb'))
@click.argument('out', type=click.File('wb'), required=False)
@click.option('--color/--no-color', default=True)
@click.option('--offline', is_flag=True)
def restore_tss(device, ipsw, out, color, offline):
    """ query SHSH blobs """
    lockdown = None
    irecv = None
    if isinstance(device, LockdownClient):
        lockdown = device
    elif isinstance(device, IRecv):
        irecv = device

    device = Device(lockdown=lockdown, irecv=irecv)
    tss = Recovery(ipsw, device, offline=offline).fetch_tss_record()
    if out:
        plistlib.dump(tss, out)
    print_json(tss, colored=color)


@restore.command('ramdisk', cls=Command)
@click.argument('ipsw', type=click.File('rb'))
@click.option('--tss', type=click.File('rb'))
def restore_ramdisk(device, ipsw, tss):
    """ don't perform an actual restore. just enter the update ramdisk """
    if tss:
        tss = plistlib.load(tss)

    lockdown = None
    irecv = None
    if isinstance(device, LockdownClient):
        lockdown = device
    elif isinstance(device, IRecv):
        irecv = device
    device = Device(lockdown=lockdown, irecv=irecv)
    Recovery(ipsw, device, tss=tss).boot_ramdisk()


@restore.command('update', cls=Command)
@click.argument('ipsw', type=click.File('rb'))
@click.option('--tss', type=click.File('rb'))
@click.option('--offline', is_flag=True)
@click.option('--erase', is_flag=True)
def restore_update(device, ipsw, tss, offline, erase):
    """ perform an upgrade """
    if tss:
        tss = plistlib.load(tss)

    lockdown = None
    irecv = None
    if isinstance(device, LockdownClient):
        lockdown = device
    elif isinstance(device, IRecv):
        irecv = device
    device = Device(lockdown=lockdown, irecv=irecv)

    behavior = 'Update'
    if erase:
        behavior = 'Erase'

    try:
        Restore(ipsw, device, tss=tss, offline=offline, behavior=behavior).update()
    except Exception:
        # click may "swallow" several exception types so we try to catch them all here
        traceback.print_exc()
        raise



from re import sub
import plistlib
import logging
import struct
import ssl

import IPython
from pygments import highlight, lexers, formatters

from pymobiledevice3 import usbmux
from pymobiledevice3.exceptions import ConnectionFailedError, PyMobileDevice3Exception

SHELL_USAGE = """
# This shell allows you to communicate directly with every service layer behind the lockdownd daemon.

# For example, you can do the following:
client.send_plist({"Command": "DoSomething"})

# and view the reply
print(client.recv_plist())

# or just send raw message
client.send(b"hello")

# and view the result
print(client.recvall(20))
"""


class ServiceConnection(object):
    def __init__(self, socket):
        self.logger = logging.getLogger(__name__)
        self.socket = socket

    @staticmethod
    def create(udid, port):
        target_device = None
        while target_device is None:
            matching_devices = [device for device in usbmux.list_devices() if device.serial == udid]
            if len(matching_devices) == 1:
                target_device = matching_devices[0]
                break

        try:
            socket = target_device.connect(port)
        except usbmux.MuxException:
            raise ConnectionFailedError(f'Connection to device port {port} failed')

        return ServiceConnection(socket)

    def setblocking(self, blocking: bool):
        self.socket.setblocking(blocking)

    def close(self):
        self.socket.close()

    def recv(self, length=4096):
        """ socket.recv() normal behavior. attempt to receive a single chunk """
        return self.socket.recv(length)

    def sendall(self, data):
        self.socket.sendall(data)

    def send_recv_plist(self, data, endianity='>', fmt=plistlib.FMT_XML):
        self.send_plist(data, endianity=endianity, fmt=fmt)
        return self.recv_plist(endianity=endianity)

    def recvall(self, size):
        data = b''
        while len(data) < size:
            chunk = self.recv(size - len(data))
            if chunk is None or len(chunk) == 0:
                raise ConnectionAbortedError()
            data += chunk
        return data

    def recv_prefixed(self, endianity='>'):
        """ receive a data block prefixed with a u32 length field """
        size = self.recvall(4)
        if not size or len(size) != 4:
            return
        size = struct.unpack(endianity + 'L', size)[0]
        return self.recvall(size)

    def send_prefixed(self, data):
        """ send a data block prefixed with a u32 length field """
        if isinstance(data, str):
            data = data.encode()
        hdr = struct.pack('>L', len(data))
        msg = b''.join([hdr, data])
        return self.sendall(msg)

    def recv_plist(self, endianity='>'):
        payload = self.recv_prefixed(endianity=endianity)
        if not payload:
            return
        bplist_header = b'bplist00'
        xml_header = b'<?xml'
        if payload.startswith(bplist_header):
            return plistlib.loads(payload)
        elif payload.startswith(xml_header):
            # HAX lockdown HardwarePlatform with null bytes
            payload = sub(r'[^\w<>\/ \-_0-9\"\'\\=\.\?\!\+]+', '', payload.decode('utf-8')).encode('utf-8')
            return plistlib.loads(payload)
        else:
            raise PyMobileDevice3Exception(f'recv_plist invalid data: {payload[:100].hex()}')

    def send_plist(self, d, endianity='>', fmt=plistlib.FMT_XML):
        payload = plistlib.dumps(d, fmt=fmt)
        message = struct.pack(endianity + 'L', len(payload))
        return self.sendall(message + payload)

    def ssl_start(self, keyfile, certfile):
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
        context.load_cert_chain(certfile, keyfile)
        self.socket = context.wrap_socket(self.socket)

    def shell(self):
        IPython.embed(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.TerminalTrueColorFormatter(style='native')),
            user_ns={
                'client': self,
            })



from pymobiledevice3.services.remote_server import MessageAux


class Graphics:
    IDENTIFIER = 'com.apple.instruments.server.services.graphics.opengl'

    def __init__(self, dvt):
        self._channel = dvt.make_channel(self.IDENTIFIER)

    def __enter__(self):
        self._channel.startSamplingAtTimeInterval_(MessageAux().append_obj(0.0))
        self._channel.receive_plist()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._channel.stopSampling()

    def __iter__(self):
        while True:
            yield self._channel.receive_plist()



import click

from pymobiledevice3.cli.cli_common import Command
from pymobiledevice3.services.mobile_activation import MobileActivationService


@click.group()
def cli():
    """ cli """
    pass


@cli.group()
def activation():
    """ activation options """
    pass


@activation.command(cls=Command)
def state(lockdown):
    """ Get current activation state """
    print(MobileActivationService(lockdown).state)


@activation.command(cls=Command)
@click.option('--offline', is_flag=True, help='Allow to send and receive requests manually')
@click.option('--now', is_flag=True, help='when --offline is used, dont wait for next nonce cycle')
def activate(lockdown, offline, now):
    """ Activate device """
    MobileActivationService(lockdown, offline, now).activate()


@activation.command(cls=Command)
def deactivate(lockdown):
    """ Deactivate device """
    MobileActivationService(lockdown).deactivate()



#!/usr/bin/env python3
import logging
from pathlib import Path
import time
import uuid
import plistlib
from datetime import datetime
from contextlib import contextmanager

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.exceptions import PyMobileDevice3Exception, AfcFileNotFoundError, AfcException
from pymobiledevice3.services.afc import AfcService, AFC_LOCK_EX, afc_error_t, AFC_LOCK_UN
from pymobiledevice3.services.device_link import DeviceLink
from pymobiledevice3.services.installation_proxy import InstallationProxyService
from pymobiledevice3.services.notification_proxy import NotificationProxyService
from pymobiledevice3.services.springboard import SpringBoardServicesService

SUPPORTED_VERSIONS = [2.0, 2.1]
ITUNES_FILES = [
    'ApertureAlbumPrefs', 'IC-Info.sidb', 'IC-Info.sidv', 'PhotosFolderAlbums', 'PhotosFolderName',
    'PhotosFolderPrefs', 'VoiceMemos.plist', 'iPhotoAlbumPrefs', 'iTunesApplicationIDs', 'iTunesPrefs',
    'iTunesPrefs.plist'
]
NP_SYNC_WILL_START = 'com.apple.itunes-mobdev.syncWillStart'
NP_SYNC_DID_START = 'com.apple.itunes-mobdev.syncDidStart'
NP_SYNC_LOCK_REQUEST = 'com.apple.itunes-mobdev.syncLockRequest'
NP_SYNC_DID_FINISH = 'com.apple.itunes-mobdev.syncDidFinish'


class Mobilebackup2Service:
    SERVICE_NAME = 'com.apple.mobilebackup2'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.service = self.lockdown.start_service(self.SERVICE_NAME)

    def backup(self, full: bool = True, backup_directory='.', progress_callback=lambda x: None) -> None:
        """
        Backup a device.
        :param full: Whether to do a full backup. If full is True, any previous backup attempts will be discarded.
        :param backup_directory: Directory to write backup to.
        :param progress_callback: Function to be called as the backup progresses.
        The function shall receive the percentage as a parameter.
        """
        backup_directory = Path(backup_directory)
        device_directory = backup_directory / self.lockdown.identifier
        device_directory.mkdir(exist_ok=True, mode=0o755, parents=True)

        with self.device_link(backup_directory) as dl:
            notification_proxy = NotificationProxyService(self.lockdown)
            afc = AfcService(self.lockdown)

            with self._backup_lock(afc, notification_proxy):
                # Initialize Info.plist
                info_plist = self.init_mobile_backup_factory_info(afc)
                with open(device_directory / 'Info.plist', 'wb') as fd:
                    plistlib.dump(info_plist, fd)

                # Initialize Status.plist file if doesn't exist.
                status_path = device_directory / 'Status.plist'
                if full or not status_path.exists():
                    with open(device_directory / 'Status.plist', 'wb') as fd:
                        plistlib.dump({
                            'BackupState': 'new',
                            'Date': datetime.utcnow(),
                            'IsFullBackup': full,
                            'Version': '3.3',
                            'SnapshotState': 'finished',
                            'UUID': str(uuid.uuid4()).upper(),
                        }, fd, fmt=plistlib.FMT_BINARY)

                # Create Manifest.plist if doesn't exist.
                manifest_path = device_directory / 'Manifest.plist'
                if full:
                    manifest_path.unlink(missing_ok=True)
                (device_directory / 'Manifest.plist').touch()

                dl.send_process_message({'MessageName': 'Backup', 'TargetIdentifier': self.lockdown.identifier})
                dl.dl_loop(progress_callback)

    def restore(self, backup_directory='.', system: bool = False, reboot: bool = True, copy: bool = True,
                settings: bool = True, remove: bool = False, password: str = '', source: str = '',
                progress_callback=lambda x: None):
        """
        Restore a previous backup to the device.
        :param backup_directory: Path of the backup directory.
        :param system: Whether to restore system files.
        :param reboot: Reboot the device when done.
        :param copy: Create a copy of backup folder before restoring.
        :param settings: Restore device settings.
        :param remove: Remove items which aren't being restored.
        :param password: Password of the backup if it is encrypted.
        :param source: Identifier of device to restore its backup.
        :param progress_callback: Function to be called as the backup progresses.
        The function shall receive the current percentage of the progress as a parameter.
        """
        backup_directory = Path(backup_directory)
        source = source if source else self.lockdown.identifier
        self._assert_backup_exists(backup_directory, source)

        with self.device_link(backup_directory) as dl:
            notification_proxy = NotificationProxyService(self.lockdown)
            afc = AfcService(self.lockdown)

            with self._backup_lock(afc, notification_proxy):
                manifest_plist_path = backup_directory / source / 'Manifest.plist'
                with open(manifest_plist_path, 'rb') as fd:
                    manifest = plistlib.load(fd)
                is_encrypted = manifest.get('IsEncrypted', False)
                options = {
                    'RestoreShouldReboot': reboot,
                    'RestoreDontCopyBackup': copy,
                    'RestorePreserveSettings': settings,
                    'RestoreSystemFiles': system,
                    'RemoveItemsNotRestored': remove,
                }
                if is_encrypted:
                    if password:
                        options['Password'] = password
                    else:
                        self.logger.error('Backup is encrypted, please supply password.')
                        return
                dl.send_process_message({
                    'MessageName': 'Restore',
                    'TargetIdentifier': self.lockdown.identifier,
                    'SourceIdentifier': source,
                    'Options': options,
                })
                dl.dl_loop(progress_callback)

    def info(self, backup_directory='.', source: str = '') -> str:
        """
        Get information about a backup.
        :param backup_directory: Path of the backup directory.
        :param source: Identifier of device to get info about its backup.
        :return: Information about a backup.
        """
        backup_dir = Path(backup_directory)
        self._assert_backup_exists(backup_dir, source if source else self.lockdown.identifier)
        with self.device_link(backup_dir) as dl:
            message = {'MessageName': 'Info', 'TargetIdentifier': self.lockdown.identifier}
            if source:
                message['SourceIdentifier'] = source
            dl.send_process_message(message)
            result = dl.dl_loop()
        return result

    def list(self, backup_directory='.', source: str = '') -> str:
        """
        List the files in the last backup.
        :param backup_directory: Path of the backup directory.
        :param source: Identifier of device to list its backup data.
        :return: List of files and additional data about each file, all in a CSV format.
        """
        backup_dir = Path(backup_directory)
        source = source if source else self.lockdown.identifier
        self._assert_backup_exists(backup_dir, source)
        with self.device_link(backup_dir) as dl:
            dl.send_process_message({
                'MessageName': 'List', 'TargetIdentifier': self.lockdown.identifier, 'SourceIdentifier': source,
            })
            result = dl.dl_loop()
        return result

    def unback(self, backup_directory='.', password: str = '', source: str = '') -> None:
        """
        Unpack a complete backup to its device hierarchy.
        :param backup_directory: Path of the backup directory.
        :param password: Password of the backup if it is encrypted.
        :param source: Identifier of device to unpack its backup.
        """
        backup_dir = Path(backup_directory)
        self._assert_backup_exists(backup_dir, source if source else self.lockdown.identifier)
        with self.device_link(backup_dir) as dl:
            message = {'MessageName': 'Unback', 'TargetIdentifier': self.lockdown.identifier}
            if source:
                message['SourceIdentifier'] = source
            if password:
                message['Password'] = password
            dl.send_process_message(message)
            dl.dl_loop()

    def extract(self, domain_name: str, relative_path: str, backup_directory='.', password: str = '',
                source: str = '') -> None:
        """
        Extract a file from a previous backup.
        :param domain_name: File's domain name, e.g., SystemPreferencesDomain or HomeDomain.
        :param relative_path: File path.
        :param backup_directory: Path of the backup directory.
        :param password: Password of the last backup if it is encrypted.
        :param source: Identifier of device to extract file from its backup.
        """
        backup_dir = Path(backup_directory)
        self._assert_backup_exists(backup_dir, source if source else self.lockdown.identifier)
        with self.device_link(backup_dir) as dl:
            message = {
                'MessageName': 'Extract', 'TargetIdentifier': self.lockdown.identifier, 'DomainName': domain_name,
                'RelativePath': relative_path
            }
            if source:
                message['SourceIdentifier'] = source
            if password:
                message['Password'] = password
            dl.send_process_message(message)
            dl.dl_loop()

    def change_password(self, backup_directory='.', old: str = '', new: str = '') -> None:
        """
        Change backup password.
        :param backup_directory: Backups directory.
        :param old: Previous password. Omit when enabling backup encryption.
        :param new: New password. Omit when disabling backup encryption.
        """
        with self.device_link(Path(backup_directory)) as dl:
            message = {'MessageName': 'ChangePassword', 'TargetIdentifier': self.lockdown.identifier}
            if old:
                message['OldPassword'] = old
            if new:
                message['NewPassword'] = new
            dl.send_process_message(message)
            dl.dl_loop()

    def erase_device(self, backup_directory='.') -> None:
        """
        Erase the device.
        """
        with self.device_link(Path(backup_directory)) as dl:
            dl.send_process_message({'MessageName': 'EraseDevice', 'TargetIdentifier': self.lockdown.identifier})
            dl.dl_loop()

    def version_exchange(self, dl: DeviceLink, local_versions=None) -> None:
        """
        Exchange versions with the device and assert that the device supports our version of the protocol.
        :param dl: Initialized device link.
        :param local_versions: versions supported by us.
        """
        if local_versions is None:
            local_versions = SUPPORTED_VERSIONS
        dl.send_process_message({
            'MessageName': 'Hello',
            'SupportedProtocolVersions': local_versions,
        })
        reply = dl.receive_message()
        assert reply[0] == 'DLMessageProcessMessage' and reply[1]['ErrorCode'] == 0
        assert reply[1]['ProtocolVersion'] in local_versions

    def init_mobile_backup_factory_info(self, afc: AfcService):
        ip = InstallationProxyService(self.lockdown)
        sbs = SpringBoardServicesService(self.lockdown)

        root_node = self.lockdown.get_value()
        itunes_settings = self.lockdown.get_value(domain='com.apple.iTunes')
        min_itunes_version = self.lockdown.get_value('com.apple.mobile.iTunes', 'MinITunesVersion')
        app_dict = {}
        installed_apps = []
        apps = ip.browse(options={'ApplicationType': 'User'},
                         attributes=['CFBundleIdentifier', 'ApplicationSINF', 'iTunesMetadata'])
        for app in apps:
            bundle_id = app['CFBundleIdentifier']
            if bundle_id:
                installed_apps.append(bundle_id)
                if app.get('iTunesMetadata', False) and app.get('ApplicationSINF', False):
                    app_dict[bundle_id] = {
                        'ApplicationSINF': app['ApplicationSINF'],
                        'iTunesMetadata': app['iTunesMetadata'],
                        'PlaceholderIcon': sbs.get_icon_pngdata(bundle_id),
                    }

        files = {}
        for file in ITUNES_FILES:
            try:
                data_buf = afc.get_file_contents('/iTunes_Control/iTunes/' + file)
            except AfcFileNotFoundError:
                pass
            else:
                files[file] = data_buf

        ret = {
            'iTunes Version': min_itunes_version if min_itunes_version else '10.0.1',
            'iTunes Files': files,
            'Unique Identifier': self.lockdown.identifier.upper(),
            'Target Type': 'Device',
            'Target Identifier': root_node['UniqueDeviceID'],
            'Serial Number': root_node['SerialNumber'],
            'Product Version': root_node['ProductVersion'],
            'Product Type': root_node['ProductType'],
            'Installed Applications': installed_apps,
            'GUID': uuid.uuid4().bytes,
            'Display Name': root_node['DeviceName'],
            'Device Name': root_node['DeviceName'],
            'Build Version': root_node['BuildVersion'],
            'Applications': app_dict,
        }

        if 'IntegratedCircuitCardIdentity' in root_node:
            ret['ICCID'] = root_node['IntegratedCircuitCardIdentity']
        if 'InternationalMobileEquipmentIdentity' in root_node:
            ret['IMEI'] = root_node['InternationalMobileEquipmentIdentity']
        if 'MobileEquipmentIdentifier' in root_node:
            ret['MEID'] = root_node['MobileEquipmentIdentifier']
        if 'PhoneNumber' in root_node:
            ret['Phone Number'] = root_node['PhoneNumber']

        try:
            data_buf = afc.get_file_contents('/Books/iBooksData2.plist')
        except AfcFileNotFoundError:
            pass
        else:
            ret['iBooks Data 2'] = data_buf
        if itunes_settings:
            ret['iTunes Settings'] = itunes_settings
        return ret

    @contextmanager
    def _backup_lock(self, afc, notification_proxy):
        notification_proxy.notify_post(NP_SYNC_WILL_START)
        lockfile = afc.fopen('/com.apple.itunes.lock_sync', 'r+')
        if lockfile:
            notification_proxy.notify_post(NP_SYNC_LOCK_REQUEST)
            for _ in range(50):
                try:
                    afc.lock(lockfile, AFC_LOCK_EX)
                except AfcException as e:
                    if e.status == afc_error_t.OP_WOULD_BLOCK:
                        time.sleep(0.2)
                    else:
                        afc.fclose(lockfile)
                        raise
                else:
                    notification_proxy.notify_post(NP_SYNC_DID_START)
                    break
            else:  # No break, lock failed.
                afc.fclose(lockfile)
                raise PyMobileDevice3Exception('Failed to lock itunes sync file')
        try:
            yield
        finally:
            afc.lock(lockfile, AFC_LOCK_UN)
            afc.fclose(lockfile)
            notification_proxy.notify_post(NP_SYNC_DID_FINISH)

    @staticmethod
    def _assert_backup_exists(backup_directory: Path, identifier: str):
        device_directory = backup_directory / identifier
        assert (device_directory / 'Info.plist').exists()
        assert (device_directory / 'Manifest.plist').exists()
        assert (device_directory / 'Status.plist').exists()

    @contextmanager
    def device_link(self, backup_directory):
        dl = DeviceLink(self.service, backup_directory)
        dl.version_exchange()
        self.version_exchange(dl)
        try:
            yield dl
        finally:
            dl.disconnect()



#!/usr/bin/env python3
import logging
import os
import platform
import plistlib
import sys
import uuid
import datetime
from distutils.version import LooseVersion
from pathlib import Path

from pymobiledevice3 import usbmux
from pymobiledevice3.ca import ca_do_everything
from pymobiledevice3.exceptions import *
from pymobiledevice3.service_connection import ServiceConnection
from pymobiledevice3.utils import sanitize_ios_version

# we store pairing records and ssl keys in ~/.pymobiledevice3
HOMEFOLDER = Path.home() / '.pymobiledevice3'
MAXTRIES = 20

LOCKDOWN_PATH = {
    'win32': Path(os.environ.get('ALLUSERSPROFILE', ''), 'Apple', 'Lockdown'),
    'darwin': Path('/var/db/lockdown/'),
    'linux': Path('/var/lib/lockdown/'),
}


def write_home_file(filename, data):
    HOMEFOLDER.mkdir(parents=True, exist_ok=True)
    filepath = HOMEFOLDER / filename
    filepath.write_bytes(data)
    return str(filepath)


def list_devices():
    return [d.serial for d in usbmux.list_devices()]


class LockdownClient(object):
    DEFAULT_CLIENT_NAME = 'pyMobileDevice'
    SERVICE_PORT = 62078

    def __init__(self, udid=None, client_name=DEFAULT_CLIENT_NAME):
        available_udids = list_devices()
        if udid is None:
            if len(available_udids) == 0:
                raise NoDeviceConnectedError()
            udid = available_udids[0]
        else:
            if udid not in available_udids:
                raise ConnectionFailedError()

        self.logger = logging.getLogger(__name__)
        self.paired = False
        self.SessionID = None
        self.service = ServiceConnection.create(udid, self.SERVICE_PORT)
        self.host_id = self.generate_host_id()
        self.system_buid = None
        self.label = client_name
        self.pair_record = None

        if self.query_type() != 'com.apple.mobile.lockdown':
            raise IncorrectModeError()

        self.all_values = self.get_value()
        self.udid = self.all_values.get('UniqueDeviceID')
        self.unique_chip_id = self.all_values.get('UniqueChipID')
        self.device_public_key = self.all_values.get('DevicePublicKey')
        self.ios_version = self.all_values.get('ProductVersion')
        self.identifier = self.udid
        if not self.identifier:
            if self.unique_chip_id:
                self.identifier = '%x' % self.unique_chip_id
            else:
                raise PyMobileDevice3Exception('Could not get UDID or ECID, failing')

        if not self.validate_pairing():
            self.pair()
            if not self.validate_pairing():
                raise FatalPairingError()
            self.service = ServiceConnection.create(udid, self.SERVICE_PORT)

    def query_type(self):
        self.service.send_plist({'Request': 'QueryType'})
        res = self.service.recv_plist()
        return res.get('Type')

    @property
    def ecid(self):
        return self.all_values['UniqueChipID']

    @property
    def date(self):
        return datetime.datetime.fromtimestamp(self.get_value(key='TimeIntervalSince1970'))

    @property
    def preflight_info(self):
        return self.get_value(key='FirmwarePreflightInfo')

    @property
    def sanitized_ios_version(self):
        return sanitize_ios_version(self.ios_version)

    def generate_host_id(self):
        hostname = platform.node()
        hostid = uuid.uuid3(uuid.NAMESPACE_DNS, hostname)
        return str(hostid).upper()

    def enter_recovery(self):
        self.service.send_plist({'Request': 'EnterRecovery'})
        return self.service.recv_plist()

    def stop_session(self):
        if self.SessionID and self.service:
            self.service.send_plist({'Label': self.label, 'Request': 'StopSession', 'SessionID': self.SessionID})
            self.SessionID = None
            res = self.service.recv_plist()
            if not res or res.get('Result') != 'Success':
                raise CannotStopSessionError()
            return res

    def get_itunes_pairing_record(self):
        platform_type = 'linux' if not sys.platform.startswith('linux') else sys.platform
        filename = LOCKDOWN_PATH[platform_type] / f'{self.identifier}.plist'
        try:
            with open(filename, 'rb') as f:
                pair_record = plistlib.load(f)
        except (PermissionError, FileNotFoundError):
            logging.warning(f'No iTunes pairing record found for device {self.identifier}')
            return None
        return pair_record

    def get_usbmux_pairing_record(self):
        self.logger.warning('Getting pair record from usbmuxd')
        client = usbmux.PlistProtocol(usbmux.MuxConnection.create_socket())
        try:
            return client.get_pair_record(self.udid)
        except PyMobileDevice3Exception:
            return None

    def get_local_pairing_record(self):
        self.logger.warning('Looking for pymobiledevice3 pairing record')
        path = HOMEFOLDER / f'{self.identifier}.plist'
        if not path.exists():
            self.logger.error(f'No pymobiledevice3 pairing record found for device {self.identifier}')
            return None
        self.logger.warning(f'Found pymobiledevice3 pairing record for device {self.udid}')
        return plistlib.loads(path.read_bytes())

    def validate_pairing(self):
        pair_record = self.get_itunes_pairing_record()
        if pair_record is not None:
            logging.info(f'Using iTunes pair record: {self.identifier}.plist')
        elif LooseVersion(self.ios_version) >= LooseVersion('13.0'):
            pair_record = self.get_usbmux_pairing_record()
        else:
            pair_record = self.get_local_pairing_record()

        if pair_record is None:
            return False

        self.pair_record = pair_record

        cert_pem = pair_record['HostCertificate']
        private_key_pem = pair_record['HostPrivateKey']

        if LooseVersion(self.ios_version) < LooseVersion('11.0'):
            validate_pair = {'Label': self.label, 'Request': 'ValidatePair', 'PairRecord': pair_record}
            self.service.send_plist(validate_pair)
            r = self.service.recv_plist()
            if (not r) or ('Error' in r):
                self.logger.error('ValidatePair fail', validate_pair)
                return False

        self.host_id = pair_record.get('HostID', self.host_id)
        self.system_buid = pair_record.get('SystemBUID', self.system_buid)
        d = {'Label': self.label, 'Request': 'StartSession', 'HostID': self.host_id, 'SystemBUID': self.system_buid}
        self.service.send_plist(d)
        start_session = self.service.recv_plist()
        self.SessionID = start_session.get('SessionID')
        if start_session.get('EnableSessionSSL'):
            lf = b'\n'
            self.ssl_file = write_home_file(f'{self.identifier}_ssl.txt', cert_pem + lf + private_key_pem)
            self.service.ssl_start(self.ssl_file, self.ssl_file)

        self.paired = True
        return True

    def pair(self):
        device_id = [d for d in usbmux.list_devices() if d.serial == self.udid][0].devid
        self.device_public_key = self.get_value('', 'DevicePublicKey')
        if self.device_public_key == b'':
            self.logger.error('Unable to retrieve DevicePublicKey')
            self.service.close()
            raise PairingError()

        self.logger.info('Creating host key & certificate')
        cert_pem, private_key_pem, device_certificate = ca_do_everything(self.device_public_key)

        pair_record = {'DevicePublicKey': self.device_public_key,
                       'DeviceCertificate': device_certificate,
                       'HostCertificate': cert_pem,
                       'HostID': self.host_id,
                       'RootCertificate': cert_pem,
                       'SystemBUID': '30142955-444094379208051516'}

        pair = self.service.send_recv_plist({'Label': self.label, 'Request': 'Pair', 'PairRecord': pair_record})

        if pair.get('Error') == 'PasswordProtected':
            self.service.close()
            raise NotTrustedError()
        elif pair.get('Result') != 'Success' and 'EscrowBag' not in pair:
            self.logger.error(pair.get('Error'))
            self.service.close()
            raise PairingError()

        pair_record['HostPrivateKey'] = private_key_pem
        pair_record['EscrowBag'] = pair.get('EscrowBag')
        write_home_file(f'{self.identifier}.plist', plistlib.dumps(pair_record))

        record_data = plistlib.dumps(pair_record)

        client = usbmux.PlistProtocol(usbmux.MuxConnection.create_socket())
        client.save_pair_record(self.udid, device_id, record_data)

        self.paired = True

    def unpair(self):
        req = {'Label': self.label, 'Request': 'Unpair', 'PairRecord': self.pair_record}
        self.service.send_plist(req)
        return self.service.recv_plist()

    def get_value(self, domain=None, key=None):
        if isinstance(key, str) and hasattr(self, 'record') and hasattr(self.pair_record, key):
            return self.pair_record[key]

        req = {'Request': 'GetValue', 'Label': self.label}

        if domain:
            req['Domain'] = domain
        if key:
            req['Key'] = key

        self.service.send_plist(req)
        res = self.service.recv_plist()
        if res:
            r = res.get('Value')
            if hasattr(r, 'data'):
                return r.data
            return r

    def set_value(self, value, domain=None, key=None):
        req = {'Request': 'SetValue', 'Label': self.label}

        if domain:
            req['Domain'] = domain
        if key:
            req['Key'] = key

        req['Value'] = value
        self.service.send_plist(req)
        response = self.service.recv_plist()
        self.logger.debug(response)
        return response

    def get_service_connection_attributes(self, name, escrow_bag=None) -> dict:
        if not self.paired:
            self.logger.info('NotPaired')
            raise NotPairedError()

        request = {'Label': self.label, 'Request': 'StartService', 'Service': name}
        if escrow_bag is not None:
            request['EscrowBag'] = escrow_bag

        self.service.send_plist(request)
        response = self.service.recv_plist()
        if not response or response.get('Error'):
            if response.get('Error', '') == 'PasswordProtected':
                raise PasswordRequiredError(
                    'your device is protected with password, please enter password in device and try again')
            raise StartServiceError(response.get("Error"))
        return response

    def start_service(self, name, escrow_bag=None) -> ServiceConnection:
        attr = self.get_service_connection_attributes(name, escrow_bag=escrow_bag)
        service_connection = ServiceConnection.create(self.udid, attr['Port'])
        if attr.get('EnableServiceSSL', False):
            service_connection.ssl_start(self.ssl_file, self.ssl_file)
        return service_connection

    def start_developer_service(self, name, escrow_bag=None) -> ServiceConnection:
        try:
            return self.start_service(name, escrow_bag)
        except (StartServiceError, ConnectionFailedError):
            self.logger.error(
                'Failed to connect to required service. Make sure DeveloperDiskImage.dmg has been mounted. '
                'You can do so using: pymobiledevice3 mounter mount'
            )
            raise



import logging
import os

import click

from pymobiledevice3.cli.cli_common import Command
from pymobiledevice3.services.afc import AfcShell, AfcService


@click.group()
def cli():
    """ crash cli """
    pass


@cli.group()
def crash():
    """ crash report options """
    pass


@crash.command('clear', cls=Command)
def crash_clear(lockdown):
    """ clear(/remove) all crash reports """
    afc = AfcService(lockdown, service_name='com.apple.crashreportcopymobile')
    afc.rm('/', force=True)


@crash.command('pull', cls=Command)
@click.argument('out', type=click.Path(file_okay=False, dir_okay=True, exists=False))
def crash_pull(lockdown, out):
    """ pull all crash reports """
    if not os.path.exists(out):
        os.makedirs(out)

    def log(src, dst):
        logging.info(f'{src} --> {dst}')

    afc = AfcService(lockdown, service_name='com.apple.crashreportcopymobile')
    afc.pull('/', out, callback=log)


@crash.command('shell', cls=Command)
def crash_shell(lockdown):
    """ start an afc shell """
    AfcShell(lockdown=lockdown, service_name='com.apple.crashreportcopymobile').cmdloop()


@crash.command('flush', cls=Command)
def crash_mover_flush(lockdown):
    """ trigger com.apple.crashreportmover to flush all products into CrashReports directory """
    ack = b'ping\x00'
    assert ack == lockdown.start_service('com.apple.crashreportmover').recvall(len(ack))
